from collections import defaultdict
from .report import *
from .seqio import FormatError

def run_serial(reader, modifiers, filters, formatters, writers):
    try:
        n = 0
        total_bp1 = 0
        total_bp2 = 0
        
        for batch_size, batch in reader:
            n += batch_size
            result = defaultdict(lambda: [])
            for record in batch:
                reads, bp = modifiers.modify(record)
                total_bp1 += bp[0]
                total_bp2 += bp[1]
                dest = filters.filter(*reads)
                formatters.format(result, dest, *reads)
            result = dict((path, "".join(strings))
                for path, strings in result.items())
            writers.write_result(result)
        
        return (0, Summary(
            collect_process_statistics(n, total_bp1, total_bp2, modifiers, filters, formatters),
            summarize_adapters(modifiers.get_adapters()),
            modifiers.get_trimmer_classes()
        ).finish())
    
    except KeyboardInterrupt as e:
        logging.getLogger().error("Interrupted")
        return (130, None)
    except IOError as e:
        if e.errno == errno.EPIPE:
            return (1, None)
        raise
    except (FormatError, EOFError) as e:
        logging.getLogger().error("Atropos error", exc_info=True)
        return(1, None)
    finally:
        reader.close()
        writers.close()



import logging
import sys

from .seqio import UnknownFileType, BatchIterator, open_reader
from .xopen import STDOUT, STDERR, open_output

def detect(options, parser):
    from .detect import (
        summarize_contaminants, PairedDetector, KnownContaminantDetector,
        HeuristicDetector, KhmerDetector)
    from .util import enumerate_range
    
    k = options.kmer_size or 12
    n_reads = options.max_reads
    overrep_cutoff = 100
    include = options.include_contaminants or "all"
    known_contaminants = load_known_adapters(options) if include != 'unknown' else None
    batch_iterator, names = create_reader(options, parser, counter_magnitude="K")[0:2]
    
    detector = options.detector
    if not detector:
        if known_contaminants and include == 'known':
            detector = 'known'
        elif n_reads <= 50000:
            detector = 'heuristic'
        else:
            detector = 'khmer'
    
    if detector == 'known':
        logging.getLogger().debug("Detecting contaminants using the known-only algorithm")
        detector_class = KnownContaminantDetector
    elif detector == 'heuristic':
        logging.getLogger().debug("Detecting contaminants using the heuristic algorithm")
        detector_class = HeuristicDetector
    elif detector == 'khmer':
        logging.getLogger().debug("Detecting contaminants using the kmer-based algorithm")
        detector_class = KhmerDetector
    
    try:
        detector_args = dict(
            k=k, n_reads=n_reads, overrep_cutoff=overrep_cutoff,
            known_contaminants=known_contaminants)
        if options.paired:
            d = PairedDetector(detector_class, **detector_args)
        else:
            d = detector_class(**detector_args)
            names = names[0]
        
        with open_output(options.output) as o:
            print("\nDetecting adapters and other potential contaminant sequences based on "
                  "{}-mers in {} reads".format(k, n_reads), file=o)
            d.consume_all_batches(batch_iterator)
            d.summarize(o, names, include=include)
    finally:
        batch_iterator.close()

def error(options, parser):
    from .error import ErrorEstimator, PairedErrorEstimator
    
    batch_iterator, names, qualities = create_reader(options, parser, counter_magnitude="K")[0:3]
    try:
        if not qualities:
            parser.error("Cannot estimate error rate without base qualities")
        
        if options.paired:
            e = PairedErrorEstimator()
        else:
            e = ErrorEstimator()
        
        e.consume_all_batches(batch_iterator)
    finally:
        batch_iterator.close()
    
    with open_output(options.output) as o:
        e.summarize(o, names)

def trim(options, parser):
    import time
    import textwrap
    from .report import print_report
    
    params = create_atropos_params(options, parser, options.default_outfile)
    num_adapters = sum(len(a) for a in params.modifiers.get_adapters())
    
    logger = logging.getLogger()
    logger.info("Trimming %s adapter%s with at most %.1f%% errors in %s mode ...",
        num_adapters, 's' if num_adapters > 1 else '', options.error_rate * 100,
        { False: 'single-end', 'first': 'paired-end legacy', 'both': 'paired-end' }[options.paired])
    if options.paired == 'first' and (len(params.modifiers.get_modifiers(read=2)) > 0 or options.quality_cutoff):
        logger.warning('\n'.join(textwrap.wrap('WARNING: Requested read '
            'modifications are applied only to the first '
            'read since backwards compatibility mode is enabled. '
            'To modify both reads, also use any of the -A/-B/-G/-U options. '
            'Use a dummy adapter sequence when necessary: -A XXX')))
    
    start_wallclock_time = time.time()
    start_cpu_time = time.clock()
    
    if options.threads is None:
        # Run single-threaded version
        import atropos.serial
        rc, summary = atropos.serial.run_serial(*params)
    else:
        # Run multiprocessing version
        import atropos.multicore
        rc, summary = atropos.multicore.run_parallel(
            params.reader, params.modifiers, params.filters, params.formatters, params.writers,
            options.threads, options.process_timeout, options.preserve_order, options.read_queue_size,
            options.result_queue_size, options.writer_process, options.compression)
    
    if rc != 0:
        sys.exit(rc)
    
    stop_wallclock_time = time.time()
    stop_cpu_time = time.clock()
    report = print_report(
        options,
        stop_wallclock_time - start_wallclock_time,
        stop_cpu_time - start_cpu_time,
        summary,
        params.modifiers.get_trimmer_classes())

def create_reader(options, parser, counter_magnitude="M"):
    interleaved = bool(options.interleaved_input)
    input1 = options.interleaved_input if interleaved else options.input1
    input2 = qualfile = None
    if options.paired and not interleaved:
        input2 = options.input2
    else:
        qualfile = options.input2
    
    try:
        reader = open_reader(input1, file2=input2, qualfile=qualfile,
            colorspace=options.colorspace, fileformat=options.format,
            interleaved=interleaved, single_input_read=options.single_input_read)
    except (UnknownFileType, IOError) as e:
        parser.error(e)
    
    qualities = reader.delivers_qualities
    
    # Wrap reader in subsampler
    if options.subsample:
        reader = subsample(reader, options.subsample)
    
    # Wrap reader in batch iterator
    batch_size = options.batch_size or 1000
    reader = BatchIterator(reader, batch_size, options.max_reads)
    
    # Wrap iterator in progress bar
    if options.progress:
        from .progress import create_progress_reader
        reader = create_progress_reader(
            reader, options.progress, batch_size, options.max_reads,
            counter_magnitude)
    
    return (reader, (input1, input2), qualities, qualfile is not None)

# TODO: specify this externally rather than hard-coding
DEFAULT_ADAPTERS_URL = \
    "https://gist.githubusercontent.com/jdidion/c151526b8ccd84945cd31dfa4f83591d/raw/" \
    "eaa6b2d99983f5b14f75357c93c1540142436a49/sequencing_adapters.fa"

def load_known_adapters(options):
    from .adapters import AdapterCache
    cache_file = options.adapter_cache_file if options.cache_adapters else None
    adapter_cache = AdapterCache(cache_file)
    if adapter_cache.empty and options.default_adapters:
        try:
            adapter_cache.load_from_url(DEFAULT_ADAPTERS_URL)
        except:
            logging.getLogger().error(
                "Error loading adapters from URL %s", DEFAULT_ADAPTERS_URL)
    if options.known_adapter:
        for s in options.known_adapter:
            name, seq = s.split('=')
            adapter_cache.add(name, seq)
    if options.known_adapters_file:
        for f in options.known_adapters_file:
            try:
                adapter_cache.load_from_url(f)
            except:
                logging.getLogger().error(
                    "Error loading adapters from url/file %s", f)
    if options.cache_adapters:
        adapter_cache.save()
    return adapter_cache

def subsample(reader, frac):
    from random import random
    for reads in reader:
        if random() < frac:
            yield reads

from collections import namedtuple
AtroposParams = namedtuple("AtroposParams", ("reader", "modifiers", "filters", "formatters", "writers"))

def create_atropos_params(options, parser, default_outfile):
    from .adapters import AdapterParser, BACK
    from .modifiers import (
        Modifiers, AdapterCutter, InsertAdapterCutter, UnconditionalCutter,
        NextseqQualityTrimmer, QualityTrimmer, NonDirectionalBisulfiteTrimmer,
        RRBSTrimmer, SwiftBisulfiteTrimmer, MinCutter, NEndTrimmer,
        LengthTagModifier, SuffixRemover, PrefixSuffixAdder, DoubleEncoder,
        ZeroCapper, PrimerTrimmer, MergeOverlapping)
    from .filters import (
        Filters, FilterFactory, TooShortReadFilter, TooLongReadFilter,
        NContentFilter, TrimmedFilter, UntrimmedFilter, NoFilter,
        MergedReadFilter)
    from .seqio import Formatters, RestFormatter, InfoFormatter, WildcardFormatter, Writers
    from .util import RandomMatchProbability
    
    reader, input_names, qualities, has_qual_file = create_reader(options, parser)
    
    if options.adapter_max_rmp or options.aligner == 'insert':
        match_probability = RandomMatchProbability()
    
    # Create Adapters
    
    has_adapters1 = options.adapters or options.anywhere or options.front
    has_adapters2 = options.adapters2 or options.anywhere2 or options.front2
    
    adapters1 = adapters2 = []
    if has_adapters1 or has_adapters2:
        adapter_cache = load_known_adapters(options)
        parser_args = dict(
            colorspace=options.colorspace,
            max_error_rate=options.error_rate,
            min_overlap=options.overlap,
            read_wildcards=options.match_read_wildcards,
            adapter_wildcards=options.match_adapter_wildcards,
            indels=options.indels, indel_cost=options.indel_cost,
            cache=adapter_cache
        )
        if options.adapter_max_rmp:
            parser_args['match_probability'] = match_probability
            parser_args['max_rmp'] = options.adapter_max_rmp
        adapter_parser = AdapterParser(**parser_args)
        
        try:
            if has_adapters1:
                adapters1 = adapter_parser.parse_multi(
                    options.adapters, options.anywhere, options.front)
            if has_adapters2:
                adapters2 = adapter_parser.parse_multi(
                    options.adapters2, options.anywhere2, options.front2)
        except IOError as e:
            if e.errno == errno.ENOENT:
                parser.error(e)
            raise
        except ValueError as e:
            parser.error(e)
        
        if options.cache_adapters:
            adapter_cache.save()
    
    # Create Modifiers
    
    if not adapters1 and not adapters2 and not options.quality_cutoff and \
            options.nextseq_trim is None and \
            options.cut == [] and options.cut2 == [] and \
            options.cut_min == [] and options.cut_min2 == [] and \
            (options.minimum_length is None or options.minimum_length <= 0) and \
            options.maximum_length == sys.maxsize and \
            not has_qual_file and options.max_n is None and not options.trim_n:
        parser.error("You need to provide at least one adapter sequence.")
    
    if options.aligner == 'insert':
        if not adapters1 or len(adapters1) != 1 or adapters1[0].where != BACK or \
                not adapters2 or len(adapters2) != 1 or adapters2[0].where != BACK:
            parser.error("Insert aligner requires a single 3' adapter for each read")
    
    if options.debug:
        for adapter in adapters1 + adapters2:
            adapter.enable_debug()
    
    modifiers = Modifiers(options.paired)
            
    for op in options.op_order:
        if op == 'A' and (adapters1 or adapters2):
            # TODO: generalize this using some kind of factory class
            if options.aligner == 'insert':
                # Use different base probabilities if we're trimming bisulfite data.
                # TODO: this doesn't seem to help things, so commenting it out for now
                #base_probs = dict(p1=0.33, p2=0.67) if options.bisulfite else dict(p1=0.25, p2=0.75)
                modifiers.add_modifier(InsertAdapterCutter,
                    adapter1=adapters1[0], adapter2=adapters2[0], action=options.action,
                    mismatch_action=options.correct_mismatches,
                    max_insert_mismatch_frac=options.insert_match_error_rate,
                    max_adapter_mismatch_frac=options.insert_match_adapter_error_rate,
                    match_probability=match_probability,
                    insert_max_rmp=options.insert_max_rmp)
            else:
                a1_args = a2_args = None
                if adapters1:
                    a1_args = dict(adapters=adapters1, times=options.times, action=options.action)
                if adapters2:
                    a2_args = dict(adapters=adapters2, times=options.times, action=options.action)
                modifiers.add_modifier_pair(AdapterCutter, a1_args, a2_args)
        elif op == 'C' and (options.cut or options.cut2):
            modifiers.add_modifier_pair(UnconditionalCutter,
                dict(lengths=options.cut),
                dict(lengths=options.cut2)
            )
        elif op == 'G' and (options.nextseq_trim is not None):
            modifiers.add_modifier(NextseqQualityTrimmer,
                read=1, cutoff=options.nextseq_trim, base=options.quality_base)
        elif op == 'Q' and options.quality_cutoff:
            modifiers.add_modifier(QualityTrimmer,
                cutoff_front=options.quality_cutoff[0],
                cutoff_back=options.quality_cutoff[1],
                base=options.quality_base)
    
    if options.bisulfite:
        if isinstance(options.bisulfite, str):
            if "non-directional" in options.bisulfite:
                modifiers.add_modifier(NonDirectionalBisulfiteTrimmer,
                    rrbs=options.bisulfite=="non-directional-rrbs")
            elif options.bisulfite == "rrbs":
                modifiers.add_modifier(RRBSTrimmer)
            elif options.bisulfite in ("epignome", "truseq"):
                # Trimming leads to worse results
                #modifiers.add_modifier(TruSeqBisulfiteTrimmer)
                pass
            elif options.bisulfite == "swift":
                modifiers.add_modifier(SwiftBisulfiteTrimmer)
        else:
            if options.bisulfite[0]:
                modifiers.add_modifier(MinCutter, read=1, **(options.bisulfite[0]))
            if len(options.bisulfite) > 1 and options.bisulfite[1]:
                modifiers.add_modifier(MinCutter, read=2, **(options.bisulfite[1]))
    
    if options.trim_n:
        modifiers.add_modifier(NEndTrimmer)
    
    if options.cut_min or options.cut_min2:
        modifiers.add_modifier_pair(MinCutter,
            dict(lengths=options.cut_min),
            dict(lengths=options.cut_min2)
        )
    
    if options.length_tag:
        modifiers.add_modifier(LengthTagModifier, length_tag=options.length_tag)
    
    if options.strip_suffix:
        modifiers.add_modifier(SuffixRemover, suffixes=options.strip_suffix)
    
    if options.prefix or options.suffix:
        modifiers.add_modifier(PrefixSuffixAdder, prefix=options.prefix, suffix=options.suffix)
    
    if options.double_encode:
        modifiers.add_modifier(DoubleEncoder)
    
    if options.zero_cap and qualities:
        modifiers.add_modifier(ZeroCapper, quality_base=options.quality_base)
    
    if options.trim_primer:
        modifiers.add_modifier(PrimerTrimmer)
    
    if options.merge_overlapping:
        modifiers.add_modifier(MergeOverlapping,
            min_overlap=options.merge_min_overlap,
            error_rate=options.merge_error_rate,
            mismatch_action=options.correct_mismatches)
    
    # Create Filters and Formatters
    
    min_affected = 2 if options.pair_filter == 'both' else 1
    filters = Filters(FilterFactory(options.paired, min_affected))
    
    output1 = output2 = None
    interleaved = False
    if options.interleaved_output:
        output1 = options.interleaved_output
        interleaved = True
    else:
        output1 = options.output
        output2 = options.paired_output
    
    seq_formatter_args = dict(
        qualities=qualities,
        colorspace=options.colorspace,
        interleaved=interleaved
    )
    formatters = Formatters(output1, seq_formatter_args)
    force_create = []
        
    if options.merge_overlapping:
        filters.add_filter(MergedReadFilter)
        if options.merged_output:
            formatters.add_seq_formatter(MergedReadFilter, options.merged_output)
        
    if options.minimum_length is not None and options.minimum_length > 0:
        filters.add_filter(TooShortReadFilter, options.minimum_length)
        if options.too_short_output:
            formatters.add_seq_formatter(TooShortReadFilter,
                options.too_short_output, options.too_short_paired_output)

    if options.maximum_length < sys.maxsize:
        filters.add_filter(TooLongReadFilter, options.maximum_length)
        if options.too_long_output is not None:
            formatters.add_seq_formatter(TooLongReadFilter,
                options.too_long_output, options.too_long_paired_output)

    if options.max_n is not None:
        filters.add_filter(NContentFilter, options.max_n)

    if options.discard_trimmed:
        filters.add_filter(TrimmedFilter)

    if not formatters.multiplexed:
        if output1 is not None:
            formatters.add_seq_formatter(NoFilter, output1, output2)
            if output1 != STDOUT and options.writer_process:
                force_create.append(output1)
                if output2 is not None:
                    force_create.append(output2)
        elif not (options.discard_trimmed and options.untrimmed_output):
            formatters.add_seq_formatter(NoFilter, default_outfile)
            if default_outfile != STDOUT and options.writer_process:
                force_create.append(default_outfile)
    
    if options.discard_untrimmed or options.untrimmed_output:
        filters.add_filter(UntrimmedFilter)

    if not options.discard_untrimmed:
        if formatters.multiplexed:
            untrimmed = options.untrimmed_output or output1.format(name='unknown')
            formatters.add_seq_formatter(UntrimmedFilter, untrimmed)
            formatters.add_seq_formatter(NoFilter, untrimmed)
        elif options.untrimmed_output:
            formatters.add_seq_formatter(UntrimmedFilter,
                options.untrimmed_output, options.untrimmed_paired_output)

    if options.rest_file:
        formatters.add_info_formatter(RestFormatter(options.rest_file))
    if options.info_file:
        formatters.add_info_formatter(InfoFormatter(options.info_file))
    if options.wildcard_file:
        formatters.add_info_formatter(WildcardFormatter(options.wildcard_file))
    
    writers = Writers(force_create)
    
    return AtroposParams(reader, modifiers, filters, formatters, writers)



# Detect adapter sequences directly from reads based on kmer frequency.

from collections import defaultdict
import logging
import math
import os
import re
import statistics as stats
import sys
from .align import Aligner, SEMIGLOBAL
from .seqio import open_reader
from .util import reverse_complement, sequence_complexity, enumerate_range
from .xopen import open_output, xopen

# TODO: Test whether using rc=True in parse_known_contaminants is as fast
# and as accurate as testing both the forward and reverse complement
# read sequence against only the forward-orientation known contaminants.
# Also, offer an option of whether to test the reverse complement, with
# the default being false.

class Match(object):
    def __init__(self, seq_or_contam, count=0, names=None, match_frac=None, match_frac2=None):
        if isinstance(seq_or_contam, ContaminantMatcher):
            self.seq = seq_or_contam.seq
            self.count = seq_or_contam.matches
            self.names = tuple(seq_or_contam.names)
            self.known_seqs = [seq_or_contam.seq]
        else:
            self.seq = seq_or_contam
            self.count = count
            self.names = names
            self.known_seqs = None
        self.match_frac = match_frac
        self.match_frac2 = match_frac2
        self.abundance = None
    
    def __len__(self):
        return len(self.seq)
    
    def __repr__(self):
        if self.is_known:
            return "{} => {} ({}))".format(self.seq, self.names, self.known_seqs)
        else:
            return self.seq
    
    @property
    def seq_complexity(self):
        return sequence_complexity(self.seq)
    
    @property
    def count_is_frequency(self):
        return isinstance(self.count, float)
    
    def set_contaminant(self, contam, match_frac, match_frac2=None):
        self.set_known(contam.names, [contam.seq], match_frac, match_frac2)
    
    def set_known(self, names, seqs, match_frac, match_frac2=None):
        self.names = names
        self.known_seqs = seqs
        self.match_frac = match_frac
        self.match_frac2 = match_frac2
    
    @property
    def is_known(self):
        return self.known_seqs is not None
    
    def estimate_abundance(self, read_sequences):
        """
        Determine whether this match's sequence is within 'seq'
        by simple exact string comparison.
        """
        self.abundance = sum(1 for read_seq in read_sequences if self.seq in read_seq)

class ContaminantMatcher(object):
    def __init__(self, seq, names, k):
        self.seq = seq
        self.names = names
        self.kmers = set(seq[i:(i+k)] for i in range(len(seq) - k + 1))
        self.n_kmers = len(self.kmers)
        self.k = k
        self.matches = 0
        
    def match(self, seq, seqrc):
        """
        Returns (num_matches, num_contam_kmers, num_seq_kmers).
        """
        fw_kmers = set(seq[i:(i+self.k)] for i in range(len(seq) - self.k + 1))
        fw_matches = float(len(self.kmers & fw_kmers))
        
        rv_kmers = set(seqrc[i:(i+self.k)] for i in range(len(seqrc) - self.k + 1))
        rv_matches = float(len(self.kmers & rv_kmers))
        
        if fw_matches >= rv_matches:
            n_matches = fw_matches
            kmers = fw_kmers
            compare_seq = seq
        else:
            n_matches = rv_matches
            kmers = rv_kmers
            compare_seq = seqrc
        
        self.matches += n_matches
        return (n_matches / self.n_kmers, n_matches / len(kmers), compare_seq)

def create_contaminant_matchers(contaminants, k):
    return [
        ContaminantMatcher(seq, names, k)
        for seq, names in contaminants.iter_sequences()
    ]

POLY_A = re.compile('A{8,}.*|A{2,}$')

class Detector(object):
    def __init__(self, k=12, n_reads=10000, overrep_cutoff=100, known_contaminants=None):
        self.k = k
        self.n_reads = n_reads
        self.overrep_cutoff = overrep_cutoff
        self.known_contaminants = known_contaminants
        self._read_length = None
        self._read_sequences = set()
        self._matches = None
    
    def consume_all(self, reader):
        """
        Consume up to self.n_reads reads from the reader.
        """
        read = next(reader)
        self.consume_first(read)
        for read in enumerate_range(reader, 1, n_reads):
            self.consume(read)
    
    def consume_all_batches(self, batch_iterator):
        """
        Consume all reads from the specified batch_iterator.
        It is expected that the iterator was constructed
        with max_reads == n_reads.
        """
        for batch_num, (batch_size, batch) in enumerate(batch_iterator):
            if batch_num == 0:
                self.consume_first(batch[0])
                batch = batch[1:]
            for read in batch:
                self.consume(read)
        
    def consume_first(self, read):
        assert self._read_length is None
        self._read_length = len(read.sequence)
        self.consume(read)
    
    def consume(self, read):
        """
        Consumes a single read.
        """
        seq = self._filter_seq(read.sequence)
        if seq:
            self._read_sequences.add(seq)
    
    def _filter_seq(self, seq):
        if sequence_complexity(seq) <= 1.0:
            return None
        match = POLY_A.search(seq)
        if match:
            seq = seq[:match.start()]
        if len(seq) < self.k:
            return None
        return seq
    
    def matches(self, **kwargs):
        if self._matches is None or len(kwargs) > 0:
            self._filter_and_sort(**kwargs)
        return self._matches
    
    def _filter_and_sort(self, include="all", min_len=None, min_complexity=1.1, min_match_frac=0.1, limit=20):
        """
        Identify, filter, and sort contaminants.
        """
        if min_len is None:
            min_len = self.k
        
        matches = self._get_contaminants()
        
        for match in matches:
            match.estimate_abundance(self._read_sequences)
        
        def _filter(match):
            if match.count < self.min_report_freq:
                return False
            if min_len and len(match) < min_len:
                return False
            if min_complexity and match.seq_complexity < min_complexity:
                return False
            if include == 'known' and not match.is_known:
                return False
            elif include == 'unknown' and match.is_known:
                return False
            if min_match_frac and match.is_known and match.match_frac < min_match_frac:
                return False
            return True
        
        matches = list(filter(_filter, matches))
        matches.sort(key=lambda x: len(x) * math.log(x.count), reverse=True)
        
        if limit is not None:
            matches = matches[:limit]
        
        self._matches = matches
    
    def summarize(self, outstream, name=None, **kwargs):
        header = "File: {}".format(name) if name else None
        summarize_contaminants(outstream, self.matches(**kwargs), self.n_reads, header)

class PairedDetector(object):
    def __init__(self, detector_class, **kwargs):
        self.d1 = detector_class(**kwargs)
        self.d2 = detector_class(**kwargs)
    
    def consume_all(self, reader):
        read1, read2 = next(reader)
        self.d1.consume_first(read1)
        self.d2.consume_first(read2)
        for read1, read2 in reader:
            self.d1.consume(read1)
            self.d2.consume(read2)
    
    def consume_all_batches(self, batch_iterator):
        for batch_num, (batch_size, batch) in enumerate(batch_iterator):
            if batch_num == 0:
                read1, read2 = batch[0]
                self.d1.consume_first(read1)
                self.d2.consume_first(read2)
                batch = batch[1:]
            for read1, read2 in batch:
                self.d1.consume(read1)
                self.d2.consume(read2)
    
    def matches(self, **kwargs):
        return (
            self.d1.matches(**kwargs),
            self.d2.matches(**kwargs))
    
    def summarize(self, outstream, names=[None, None], **kwargs):
        name1, name2 = names
        self.d1.summarize(outstream, name1, **kwargs)
        self.d2.summarize(outstream, name2, **kwargs)

class KnownContaminantDetector(Detector):
    """
    Test known contaminants against reads.
    This has linear complexity and is more specific than the khmer matcher, but
    less specific than the heuristic matcher. It's also less sensitive since
    it does not try to detect unknown contaminants.
    """
    def __init__(self, known_contaminants, min_match_frac=0.5, **kwargs):
        super(KnownContaminantDetector, self).__init__(known_contaminants=known_contaminants, **kwargs)
        self.min_match_frac = min_match_frac
        self._min_k = min(len(s) for s in known_contaminants.sequences)
    
    @property
    def min_report_freq(self):
        return 0.1
    
    def _filter_seq(self, seq):
        seq = super(KnownContaminantDetector, self)._filter_seq(seq)
        if seq and len(seq) >= self._min_k:
            return seq
        return None
    
    def _get_contaminants(self, read_seqs):
        contaminant_matchers = create_contaminant_matchers(known_contaminants, k)
        counts = defaultdict(lambda: 0)

        for seq in read_seqs:
            seqrc = reverse_complement(seq)
            for contam in contaminant_matchers:
                match = contam.match(seq, seqrc)
                if match[0] > self.min_match_frac:
                    counts[contam] += 1
        
        min_count = math.ceil(
            self.n_reads * (self._read_length - self._min_k + 1) * self.overrep_cutoff /
            float(4**self._min_k))
        
        return [
            Match(c[0], match_frac=float(c[1]) / self.n_reads)
            for c in filter(
                lambda x: x[1] >= min_count,
                counts.items()
            )
        ]

class HeuristicDetector(Detector):
    """
    Use a heuristic iterative algorithm to arrive at likely contaminants.
    This is the most accurate algorithm overall, but it has quadratic complexity
    and becomes too slow/memory-intenstive when n_reads > 50k.
    """
    def __init__(self, min_freq=0.001, min_contaminant_match_frac=0.9, **kwargs):
        super(HeuristicDetector, self).__init__(**kwargs)
        self.min_freq = min_freq
        self.min_contaminant_match_frac = min_contaminant_match_frac
    
    @property
    def min_report_freq(self):
        return 0.1 * self.n_reads
    
    def _get_contaminants(self):
        def _min_count(k):
            return math.ceil(self.n_reads * max(
                self.min_freq,
                (self._read_length - k + 1) * self.overrep_cutoff / float(4**k)))
        
        k = self.k
        kmers = defaultdict(lambda: [0, set()])
        
        for seq in self._read_sequences:
            for i in range(len(seq) - k + 1):
                kmer = seq[i:(i+k)]
                kmers[kmer][0] += 1
                kmers[kmer][1].add(seq)
        
        prev = None
        cur = {}
        results = {}
        min_count = _min_count(k)
        
        # Identify candidate kmers for increasing values of k
        while True:
            all_seqs = set()
            for kmer, (count, seqs) in kmers.items():
                if count > min_count:
                    cur[kmer] = (count, seqs)
                    all_seqs.update(seqs)
            
            if len(all_seqs) == 0:
                break
            
            if prev:
                for kmer, (count, seqs) in prev.items():
                    if not any(seq in cur for seq in seqs) and sequence_complexity(kmer) > 1.0:
                        results[kmer] = count
            
            k += 1
            kmers = defaultdict(lambda: [0, set()])
            for seq in all_seqs:
                for i in range(len(seq) - k + 1):
                    kmer = seq[i:(i+k)]
                    kmers[kmer][0] += 1
                    kmers[kmer][1].add(seq)
            
            min_count = _min_count(k)
            prev = cur
            cur = {}
        
        results = list(results.items())
        
        # Now merge overlapping sequences by length and frequency to eliminate
        # redundancy in the set of candidate kmers.
        results.sort(key=lambda i: len(i[0]) * math.log(i[1]), reverse=True)
        cur = results[0]
        merged = []
        unmerged = []
        while len(results) > 1:
            seq1, count1 = results[0]
            for j in range(1, len(results)):
                seq2, count2 = results[j]
                if len(seq1) >= len(seq2) and seq2 in seq1:
                    count1 += count2
                elif seq1 in seq2:
                    # if they are close in count, keep the longer sequence
                    if count1 < (2 * count2):
                        seq1 = seq2
                    count1 += count2
                else:
                    unmerged.append(results[j])
            merged.append([seq1, count1])
            results = unmerged
            unmerged = []
        results = merged + results
        
        if len(results) == 0:
            return []
            
        # Re-sort by frequency
        results.sort(key=lambda i: i[1], reverse=True)
        # Keep anything that's within 50% of the top hit
        # TODO: make this user-configurable?
        min_count = int(results[0][1] * 0.5)
        results = filter(lambda x: x[1] >= min_count, results)
        # Convert to matches
        matches = [Match(x[0], x[1]) for x in results]
        
        if self.known_contaminants:
            # Match to known sequences
            contaminants = create_contaminant_matchers(self.known_contaminants, self.k)
            known = {}
            unknown = []
            
            for match in matches:
                seq = match.seq
                seqrc = reverse_complement(seq)
                best_matches = {}
                best_match_frac = (self.min_contaminant_match_frac, 0)
                for contam in contaminants:
                    match_frac1, match_frac2, compare_seq = contam.match(seq, seqrc)
                    if match_frac1 < best_match_frac[0]:
                        continue
                    if (contam.seq in compare_seq or
                            align(compare_seq, contam.seq, self.min_contaminant_match_frac)):
                        if (match_frac1 > best_match_frac[0] or (
                                match_frac1 == best_match_frac[0] and
                                match_frac2 > best_match_frac[1])):
                            best_matches = {}
                            best_match_frac = (match_frac1, match_frac2)
                        best_matches[contam] = (match, (match_frac1, match_frac2))
                
                if best_matches:
                    for contam, match in best_matches.items():
                        if contam not in known or match[1] > known[contam][1]:
                            known[contam] = match
                else:
                    unknown.append(match)
            
            # resolve many-many relationships
            
            new_matches = defaultdict(lambda: [])
            for contam, (match, match_frac) in known.items():
                new_matches[match].append((contam, match_frac))
            
            known = []
            for match, contams in new_matches.items():
                if len(contams) == 1:
                    contam, match_frac = contams[0]
                    match.set_contaminant(contam, *match_frac)
                else:
                    contams.sort(key=lambda x: x[1], reverse=True)
                    contam, match_frac = contams[0]
                    equiv = [c for c in contams[1:] if c[1] == match_frac]
                    if len(equiv) == 0:
                        match.set_contaminant(contam, *match_frac)
                    else:
                        names = set(contam.names)
                        seqs = set((contam.seq,))
                        for e in equiv:
                            names.update(e[0].names)
                            seqs.add(e[0].seq)
                        match.set_known(list(names), list(seqs), *match_frac)
                known.append(match)
            
            matches = known + unknown
        
        return matches

class KhmerDetector(Detector):
    """
    Identify contaminants based on kmer frequency using a fast kmer counting
    approach (as implemented in the khmer library). This approach is fast but
    not as accurate as the other two.
    """
    def __init__(self, **kwargs):
        # make sure khmer is installed
        import khmer
        super(KhmerDetector, self).__init__(**kwargs)
        
    @property
    def min_report_freq(self):
        return 0.0001
    
    def _get_contaminants(self):
        from khmer import khmer_args
        n_win = self._read_length - self.k + 1 # assuming all sequences are same length
        tablesize = n_reads * n_win
        countgraph = khmer.Countgraph(self.k, tablesize, khmer_args.DEFAULT_N_TABLES)
        countgraph.set_use_bigcount(True)
        
        for seq in self._read_sequences:
            countgraph.consume_and_tag(seq)
        
        n_expected = math.ceil(tablesize / float(4**k))
        min_count = n_expected * self.overrep_cutoff
        if min_count >= 2**16:
            raise Exception("The minimum count for an over-represented k-kmer {} "
                            "is greater than the max khmer count (2^16)".format(min_count))
    
        candidates = {}
        
        for tag in countgraph.get_tagset():
            count = countgraph.get(tag)
            if count >= min_count:
                candidates[tag] = count
        
        if self.known_contaminants:
            matches = []
            seen = set()
            
            def match(kmer):
                n = candidates.get(kmer, 0)
                if n > 0:
                    seen.add(kmer)
                return n
            
            for seq, names in self.known_contaminants.iter_sequences():
                l = len(seq)
                if l < k:
                    print("Cannot check {}; sequence is shorter than {}".format(list(names)[0], k))
                    continue
                
                n_kmers = l - self.k + 1
                num_matches = 0
                match_counts = []
                for i in range(n_kmers):
                    kmer = seq[i:(i+k)]
                    kmer_count = max(
                        match(kmer),
                        match(reverse_complement(kmer))
                    )
                    if kmer_count > 0:
                        num_matches += 1
                        match_counts.append(kmer_count)
                
                if num_matches > 0:
                    # not sure what the correct metric is to use here
                    overall_count = sum(match_counts) / float(n_kmers)
                    matches.append(Match(
                        seq, overall_count / float(tablesize), names, float(num_matches) / n_kmers))
            
            # Add remaining tags
            for tag in set(candidates.keys()) - seen:
                matches.append(Match(tag, candidates[tag] / float(tablesize)))

        else:
            matches = [
                Match(tag, count / float(tablesize))
                for tag, count in candidates.items()
            ]
        
        return matches

def align(seq1, seq2, min_overlap_frac=0.9):
    aligner = Aligner(
        seq1, 0.0,
        SEMIGLOBAL,
        False, False)
    aligner.min_overlap = math.ceil(min(len(seq1), len(seq2)) * min_overlap_frac)
    aligner.indel_cost = 100000
    match = aligner.locate(seq2)
    if match:
        return seq1[match[0]:match[1]]
    else:
        return None

def summarize_contaminants(outstream, matches, n_reads, header=None):
    print("", file=outstream)
    
    if header:
        print(header, file=outstream)
        print('-' * len(header), file=outstream)
    
    print("Detected {} adapters/contaminants:".format(len(matches)), file=outstream)
    
    pad = len(str(len(matches)))
    for i, match in enumerate(matches):
        print(("{:>" + str(pad) + "}. {}").format(i+1, match.seq), file=outstream)
        if match.is_known:
            print("    Name(s): {}".format(",\n             ".join(match.names)), file=outstream)
            print("    Known sequence(s): {}".format(",\n             ".join(match.known_seqs)), file=outstream)
            print("    Known sequence K-mers that match detected contaminant: {:.2%}".format(match.match_frac), file=outstream)
        if match.abundance:
            print("    Abundance (full-length) in {} reads: {} ({:.1%})".format(n_reads, match.abundance, match.abundance / n_reads))
        if match.match_frac2:
            print("    Detected contaminant kmers that match known sequence: {:.2%}".format(match.match_frac2), file=outstream)
        if match.count_is_frequency:
            print("    Frequency of k-mers: {:.2%}".format(match.count), file=outstream)
        else:
            print("    Number of k-mer matches: {}".format(match.count), file=outstream)



"""
Open compressed files transparently.
"""
import errno
import io
import os
import sys

from .compression import get_file_opener

STDOUT = '-'
STDERR = '_'

def abspath(path):
    return os.path.abspath(os.path.expanduser(path))
    
def resolve_path(path, parent=None):
    """Resolves the absolute path of the specified file.
    
    Args:
        path (str): Path to resolve.
        parent (str): The directory containing ``path`` if ``path`` is relative.
    
    Returns:
        The absolute path.
    
    Raises:
        IOError: if the path does not exist.
    """
    apath = abspath(path)
    if not os.path.exists(apath) and parent is not None:
        apath = abspath(os.path.join(parent, path))
    if not os.path.exists(apath):
        raise IOError(errno.ENOENT, "%s does not exist" % apath, apath)
    return apath

def check_path(path, ptype=None, access=None):
    """Checks that a path exists, is of the specified type, and allows the specified access.
    
    Args:
        ptype: 'f' for file or 'd' for directory.
        access (int): One of the access values from :module:`os`
    
    Raises:
        IOError if the path does not exist, is not of the specified type, or doesn't allow the
        specified access.
    """
    if ptype == 'f' and not path.startswith("/dev/") and not os.path.isfile(path):
        raise IOError(errno.EISDIR, "{} is not a file".format(path), path)
    elif ptype == 'd' and not os.path.isdir(path):
        raise IOError(errno.ENOTDIR, "{} is not a directory".format(path), path)
    elif not os.path.exists(path):
        raise IOError(errno.ENOENT, "{} does not exist".format(path), path)
    if access is not None and not os.access(path, access):
        raise IOError(errno.EACCES, "{} is not accessable".format(path), path)
    return path

def check_writeable(p, ptype=None):
    if p in (STDOUT, STDERR):
        return p
    p = abspath(p)
    try:
        path = resolve_path(p)
        check_path(path, ptype, os.W_OK)
    except IOError:
        dirpath = os.path.dirname(p)
        if os.path.exists(dirpath):
            check_path(dirpath, "d", os.W_OK)
        else:
            os.makedirs(dirpath)
        path = os.path.join(dirpath, os.path.basename(p))
    return path

def open_output(filename, mode='w', context_wrapper=False):
    """
    Replacement for the "open" function that is only for writing text files.
    If the filename is '-', standard output (mode 'w').
    
    mode can be: 'a', 'ab', 'wt', or 'wb'
    Instead of 'wt', 'w' can be used as an abbreviation.
    """
    if mode == 'w':
        mode = 'wt'
    elif mode == 'a':
        mode = 'at'
    if mode not in ('wt', 'wb', 'at', 'ab'):
        raise ValueError("mode '{0}' not supported".format(mode))
    if not isinstance(filename, str):
        raise ValueError("the filename must be a string")

    # standard input and standard output handling
    if filename in (STDOUT, STDERR):
        fh = sys.stdout if filename == STDOUT else sys.stderr
        if mode == 'wb':
            fh = fh.buffer
        if context_wrapper:
            class StdWrapper(object):
                def __init__(self, fh):
                    self.fh = fh
                def __enter__(self):
                    return self.fh
                def __exit__(exception_type, exception_value, traceback):
                    pass
            fh = StdWrapper(fh)
    else:
        filename = check_writeable(filename, 'f')
        fh = open(filename, mode)
    
    return fh

def xopen(filename, mode='r', use_system=True):
    """
    Replacement for the "open" function that can also open files that have
    been compressed with gzip, bzip2 or xz. If the filename is '-', standard
    output (mode 'w') or input (mode 'r') is returned. If the filename ends
    with .gz, the file is opened with a pipe to the gzip program. If that
    does not work, then gzip.open() is used (the gzip module is slower than
    the pipe to the gzip program). If the filename ends with .bz2, it's
    opened as a bz2.BZ2File. Otherwise, the regular open() is used.

    mode can be: 'rt', 'rb', 'a', 'wt', or 'wb'
    Instead of 'rt' and 'wt', 'r' and 'w' can be used as abbreviations.

    Append mode ('a') is unavailable with BZ2 compression and will raise an error.
    """
    if mode == 'r':
        mode = 'rt'
    elif mode == 'w':
        mode = 'wt'
    elif mode == 'a':
        mode = 'at'
    if mode not in ('rt', 'rb', 'wt', 'wb', 'at', 'ab'):
        raise ValueError("mode '{0}' not supported".format(mode))
    if not isinstance(filename, str):
        raise ValueError("the filename must be a string")

    # standard input and standard output handling
    if filename in (STDOUT, STDERR):
        if 'r' in mode:
            fh = sys.stdin
        else:
            fh = sys.stdout if filename == STDOUT else sys.stderr
        if 'b' in mode:
            fh = fh.buffer
        return fh
    
    file_opener = get_file_opener(filename)
    if file_opener:
        return file_opener(filename, mode, use_system)
    else:
         return open(filename, mode)



import io
import os
import sys
from subprocess import Popen, PIPE

# Inline compression

import gzip
import bz2
import lzma

compressors = {
    ".gz"  : gzip,
    ".bz2" : bz2,
    ".xz"  : lzma
}

def get_compressor(filename):
    ext = os.path.splitext(filename)[1]
    if ext in compressors:
        return compressors[ext]
    return None

# File compression
# Using a system process can be faster, but only works
# on a system with gzip installed.

class GzipWriter:
    def __init__(self, path, mode='w'):
        self.name = path
        self.outfile = open(path, mode)
        self.devnull = open(os.devnull, 'w')
        self.closed = False
        try:
            # Setting close_fds to True is necessary due to
            # http://bugs.python.org/issue12786
            self.process = Popen([get_program_path('gzip')], stdin=PIPE, stdout=self.outfile,
                stderr=self.devnull, close_fds=True)
        except IOError as e:
            self.outfile.close()
            self.devnull.close()
            raise
    
    def readable(self): return False
    def writable(self): return True
    def seekable(self): return False
    
    def write(self, arg):
        self.process.stdin.write(arg)
    
    def flush(self):
        self.process.stdin.flush()
    
    def close(self):
        self.closed = True
        self.process.stdin.close()
        retcode = self.process.wait()
        self.outfile.close()
        self.devnull.close()
        if retcode != 0:
            raise IOError("Output gzip process terminated with exit code {0}".format(retcode))

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()

class GzipReader:
    def __init__(self, path):
        self.name = path
        self.process = Popen([get_program_path('gzip'), '-cd', path], stdout=PIPE)
        self.closed = False
    
    def readable(self): return True
    def writable(self): return False
    def seekable(self): return False
    def flush(self): pass
    
    def close(self):
        self.close = True
        retcode = self.process.poll()
        if retcode is None:
            # still running
            self.process.terminate()
        self._raise_if_error()

    def __iter__(self):
        for line in self.process.stdout:
            yield line
        self.process.wait()
        self._raise_if_error()

    def _raise_if_error(self):
        """
        Raise EOFError if process is not running anymore and the
        exit code is nonzero.
        """
        retcode = self.process.poll()
        if retcode is not None and retcode != 0:
            raise EOFError("gzip process returned non-zero exit code {0}. Is the "
                "input file truncated or corrupt?".format(retcode))

    def read(self, *args):
        data = self.process.stdout.read(*args)
        if len(args) == 0 or args[0] <= 0:
            # wait for process to terminate until we check the exit code
            self.process.wait()
        self._raise_if_error()
        return data

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()

def open_gzip_file(filename, mode, use_system=True):
    if use_system:
        try:
            if 'r' in mode:
                z = GzipReader(filename)
            else:
                z = GzipWriter(filename)
            if 't' in mode:
                z = io.TextIOWrapper(z)
            return z
        except:
            pass
    
    z = gzip.open(filename, mode)
    if 'b' in mode:
        if 'r' in mode:
            z = io.BufferedReader(z)
        else:
            z = io.BufferedWriter(z)
    return z

# TODO: experiment with system process-based bzip and lzma readers/writers

def open_bzip_file(filename, mode, use_system=False):
    if 't' in mode:
        return io.TextIOWrapper(bz2.BZ2File(filename, mode[0]))
    else:
        return bz2.BZ2File(filename, mode)

def open_lzma_file(filename, mode, use_system=False):
    return lzma.open(filename, mode)

file_openers = {
    ".gz"  : open_gzip_file,
    ".bz2" : open_bzip_file,
    ".xz"  : open_lzma_file,
}

def can_use_system_compression():
    return get_program_path("gzip") is not None

def get_file_opener(filename):
    ext = os.path.splitext(filename)[1]
    if ext in file_openers:
        return file_openers[ext]
    return None

program_cache = {}

def get_program_path(program):
    if program in program_cache:
        return program_cache[program]
    
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)
    
    exe_file = None
    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            exe_file = program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                break
    
    program_cache[program] = exe_file
    return exe_file

def open_compressed_file(filename, mode):
    ext = os.path.splitext(filename)
    opener = get_file_opener(ext)
    if not opener:
        raise Exception("{} is not a recognized compression format")
    return opener(filename, mode)



# coding: utf-8
"""
This module implements all the read modifications that atropos supports.
A modifier must be callable. It is implemented as a function if no parameters
need to be stored, and as a class with a __call__ method if there are parameters
(or statistics).
"""
import copy
import logging
import re
from .qualtrim import quality_trim_index, nextseq_trim_index
from .align import Aligner, InsertAligner, SEMIGLOBAL, START_WITHIN_SEQ1, STOP_WITHIN_SEQ2
from .util import base_complements, reverse_complement, mean

# Base classes

class ReadPairModifier(object):
    """Superclass of modifiers that edit a pair of reads simultaneously."""
    def __call__(self, read1, read2):
        raise NotImplemented()

class Trimmer(object):
    def __init__(self):
        self.trimmed_bases = 0
    
    def __call__(self, read):
        raise NotImplemented()
    
    def subseq(self, read, begin=0, end=None):
        if begin or (end is not None):
            front_bases, back_bases, new_read = read.subseq(begin, end)
            self.trimmed_bases += front_bases + back_bases
            return new_read
        else:
            return read
    
    def clip(self, read, front=0, back=0):
        if (front or back) and len(read) > 0:
            front_bases, back_bases, new_read = read.clip(front, back)
            self.trimmed_bases += front_bases + back_bases
            return new_read
        else:
            return read

# Modifiers

class AdapterCutter(object):
    """
    Repeatedly find one of multiple adapters in reads.
    The number of times the search is repeated is specified by the
    times parameter.
    """

    def __init__(self, adapters=[], times=1, action='trim'):
        """
        adapters -- list of Adapter objects

        action -- What to do with a found adapter: None, 'trim', or 'mask'
        """
        super(AdapterCutter, self).__init__()
        self.adapters = adapters
        self.times = times
        self.action = action
        self.with_adapters = 0

    def _best_match(self, read):
        """
        Find the best matching adapter in the given read.

        Return either a Match instance or None if there are no matches.
        """
        best = None
        for adapter in self.adapters:
            match = adapter.match_to(read)
            if match is None:
                continue

            # the no. of matches determines which adapter fits best
            if best is None or match.matches > best.matches:
                best = match
        return best

    def __call__(self, read):
        """
        Determine the adapter that best matches the given read.
        Since the best adapter is searched repeatedly, a list
        of Match instances is returned, which
        need to be applied consecutively to the read.
        The list is empty if there are no adapter matches.

        The read is converted to uppercase before it is compared to the adapter
        sequences.

        Cut found adapters from a single read. Return modified read.
        """
        if len(read) == 0:
            return read
        
        matches = []

        # try at most self.times times to remove an adapter
        trimmed_read = read
        for t in range(self.times):
            match = self._best_match(trimmed_read)
            if match is None:
                # nothing found
                break
            matches.append(match)
            trimmed_read = match.adapter.trimmed(match)
        
        if not matches:
            trimmed_read.match = None
            trimmed_read.match_info = None
            return trimmed_read
        
        if __debug__:
            assert len(trimmed_read) < len(read), "Trimmed read isn't shorter than original"
        
        if self.action == 'trim':
            # read is already trimmed, nothing to do
            pass
        elif self.action == 'mask':
            # add N from last modification
            masked_sequence = trimmed_read.sequence
            for match in sorted(matches, reverse=True, key=lambda m: m.astart):
                ns = 'N' * (len(match.read.sequence) -
                            len(match.adapter.trimmed(match).sequence))
                # add N depending on match position
                if match.front:
                    masked_sequence = ns + masked_sequence
                else:
                    masked_sequence += ns
            # set masked sequence as sequence with original quality
            trimmed_read.sequence = masked_sequence
            trimmed_read.qualities = matches[0].read.qualities

            assert len(trimmed_read.sequence) == len(read)
        elif self.action is None:
            trimmed_read = read
        
        trimmed_read.match = matches[-1]
        trimmed_read.match_info = [match.get_info_record() for match in matches]
        
        self.with_adapters += 1
        return trimmed_read

# TODO: For bases in overlapping sequences, should we also correct
# the base qualities to be the highest of the two?

class ErrorCorrectorMixin(object):
    """Provides a method for error correction.
    
    Args:
        mismatch_action: The action to take when a mismatch between the
            overlapping portions of read1 and read2 is encountered. Valid
            values are 'liberal', 'conservative', 'N'.
        min_qual_difference: When mismatch_action=='conservative', the minimum
            difference in base quality between read1 and read2 required to
            perform the correction.
    """
    def __init__(self, mismatch_action=None, min_qual_difference=1):
        self.mismatch_action = mismatch_action
        self.r1r2_min_qual_difference = min_qual_difference
        self.r2r1_min_qual_difference = -1 * min_qual_difference
        self.corrected_pairs = 0
        self.corrected_bp = [0, 0]
    
    def correct_errors(self, read1, read2, insert_match):
        # read2 reverse-complement is the reference, read1 is the query
        r1_seq = list(read1.sequence)
        r2_seq = list(read2.sequence)
        l2 = len(r2_seq)
        
        has_quals = read1.qualities and read2.qualities
        if has_quals:
            r1_qual = list(read1.qualities)
            r2_qual = list(read2.qualities)
        elif self.mismatch_action in ('liberal', 'conservative'):
            raise Exception("Cannot perform quality-based error correction on "
                            "reads lacking quality information")
        
        r1_start = insert_match[2]
        r1_end = insert_match[3]
        r1_changed = 0
        r2_start = l2 - insert_match[1]
        r2_end = l2 - insert_match[0]
        r2_changed = 0
        quals_equal = []
        
        for i, j in zip(range(r1_start, r1_end), range(r2_end - 1, r2_start, -1)):
            b1 = r1_seq[i]
            b2 = base_complements[r2_seq[j]]
            if b1 == b2:
                continue
            if self.mismatch_action == 'N':
                r1_seq[i] = 'N'
                r2_seq[j] = 'N'
                r1_changed += 1
                r2_changed += 1
            elif b1 == 'N':
                r1_seq[i] = b2
                if has_quals:
                    r1_qual[i] = r2_qual[j]
                r1_changed += 1
            elif b2 == 'N':
                r2_seq[j] = base_complements[b1]
                if has_quals:
                    r2_qual[j] = r1_qual[i]
                r2_changed += 1
            elif has_quals:
                diff = ord(r1_qual[i]) - ord(r2_qual[j])
                if diff >= self.r1r2_min_qual_difference:
                    r2_seq[j] = base_complements[b1]
                    r2_qual[j] = r1_qual[i]
                    r2_changed += 1
                elif diff <= self.r2r1_min_qual_difference:
                    r1_seq[i] = b2
                    r1_qual[i] = r2_qual[j]
                    r1_changed += 1
                elif self.mismatch_action == 'liberal':
                    quals_equal.append((i, j, b1, b2))
        
        if quals_equal:
            mean_qual1 = mean([ord(b) for b in r1_qual[r1_start:r1_end]])
            mean_qual2 = mean([ord(b) for b in r2_qual[r2_start:r2_end]])
            # Only make the corrections if one read is significantly better
            # than the other.
            # TODO: this method of determining whether one read is better
            # than the other is crude - come up with something better.
            diff = mean_qual1 - mean_qual2
            if diff > 1:
                # read1 is better than read2
                for i, j, b1, b2 in quals_equal:
                    r2_seq[j] = base_complements[b1]
                    r2_qual[j] = r1_qual[i]
                    r2_changed += 1
            elif diff < -1:
                # read2 is better than read1
                for i, j, b1, b2 in quals_equal:
                    r1_seq[i] = b2
                    r1_qual[i] = r2_qual[j]
                    r1_changed += 1
        
        if r1_changed or r2_changed:
            self.corrected_pairs += 1
            if r1_changed:
                self.corrected_bp[0] += r1_changed
                read1.sequence = ''.join(r1_seq)
                read1.corrected = r1_changed
                if has_quals:
                    read1.qualities = ''.join(r1_qual)
            if r2_changed:
                self.corrected_bp[1] += r2_changed
                read2.sequence = ''.join(r2_seq)
                read2.corrected = r2_changed
                if has_quals:
                    read2.qualities = ''.join(r2_qual)

class InsertAdapterCutter(ReadPairModifier, ErrorCorrectorMixin):
    """
    AdapterCutter that uses InsertAligner to first try to identify
    insert overlap before falling back to semi-global adapter alignment.
    """
    def __init__(self, adapter1, adapter2, action='trim', mismatch_action=None,
                 symmetric=True, min_insert_overlap=1, **aligner_args):
        ErrorCorrectorMixin.__init__(self, mismatch_action)
        self.adapter1 = adapter1
        self.adapter2 = adapter2
        self.aligner = InsertAligner(adapter1.sequence, adapter2.sequence,
            min_insert_overlap=min_insert_overlap, **aligner_args)
        self.min_insert_len = min_insert_overlap
        self.action = action
        self.symmetric = symmetric
        self.with_adapters = [0, 0]
    
    def __call__(self, read1, read2):
        read_lengths = [len(r) for r in (read1, read2)]
        if any(l < self.min_insert_len for l in read_lengths):
            return (read1, read2)
        
        match = self.aligner.match_insert(read1.sequence, read2.sequence)
        read1.insert_overlap = read2.insert_overlap = (match is not None)
        insert_match = None
        correct_errors = False
        
        if match:
            insert_match, adapter_match1, adapter_match2 = match
            correct_errors = self.mismatch_action and insert_match[5] > 0
        else:
            adapter_match1 = self.adapter1.match_to(read1)
            adapter_match2 = self.adapter2.match_to(read2)
            # If the adapter matches are complementary, perform error correction
            if (self.mismatch_action and adapter_match1 and adapter_match2 and
                        adapter_match1.rstart == adapter_match2.rstart):
                    insert_match = (
                        read_lengths[1] - adapter_match1.rstart,
                        read_lengths[1], 0, adapter_match1.rstart)
                    correct_errors = True
        
        # If exactly one of the two alignments failed and symmetrix is True,
        # duplicate the good alignment
        if self.symmetric and sum(
                bool(m) for m in (adapter_match1, adapter_match2)) == 1:
            if adapter_match1:
                adapter_match2 = adapter_match1.copy()
            else:
                adapter_match1 = adapter_match2.copy()
            if self.mismatch_action and not insert_match:
                # Assume that the symmetric read segments overlap and
                # perform error correction
                insert_match = (
                    read_lengths[1] - adapter_match1.rstart,
                    read_lengths[1], 0, adapter_match1.rstart)
                correct_errors = True
        
        if correct_errors:
            self.correct_errors(read1, read2, insert_match)
        
        return (
            self.trim(read1, self.adapter1, adapter_match1, 0),
            self.trim(read2, self.adapter2, adapter_match2, 1)
        )
    
    def trim(self, read, adapter, match, read_idx):
        if not match:
            read.match = None
            read.match_info = None
            return read
        
        match.adapter = adapter
        match.read = read
        match.front = False
    
        if self.action is None or match.rstart >= len(read):
            trimmed_read = read
        
        else:
            trimmed_read = adapter.trimmed(match)
            
            if self.action == 'mask':
                # add N from last modification
                masked_sequence = trimmed_read.sequence
                masked_sequence += ('N' * len(read) - len(trimmed_read))
                # set masked sequence as sequence with original quality
                trimmed_read.sequence = masked_sequence
                trimmed_read.qualities = read.qualities
            elif self.action == 'lower':
                # TODO: offer option to mask with lower-case of trimmed base
                # This will happen as part of the refactoring to modify
                # Sequences in-place.
                pass
        
        trimmed_read.match = match
        trimmed_read.match_info = [match.get_info_record()]
        
        self.with_adapters[read_idx] += 1
        return trimmed_read

class UnconditionalCutter(Trimmer):
    """
    A modifier that unconditionally removes the first n or the last n bases from a read.
    Equivalent to MinCutter with count_trimmed=False and only_trimmed=False, but
    UnconditionalCutter is applied before adapter trimming.
    
    If the length is positive, the bases are removed from the beginning of the read.
    If the length is negative, the bases are removed from the end of the read.
    """
    
    display_str = "Cut unconditionally"
    
    def __init__(self, lengths=[]):
        super(UnconditionalCutter, self).__init__()
        self.front_length = sum(l for l in lengths if l > 0)
        self.back_length = sum(l for l in lengths if l < 0)

    def __call__(self, read):
        return self.clip(read, self.front_length, self.back_length)

class MinCutter(Trimmer):
    """
    Ensure that a minimum number of bases have been trimmed off each end.
    count_trimmed :: whether to consider bases cut before or during adapter trimming
        when counting the number of bases that have already been cut
    only_trimmed :: only cut read ends if they have already been adapter-trimmed.
    """
    
    display_str = "Cut conditionally"
    
    def __init__(self, lengths=[], count_trimmed=True, only_trimmed=False):
        super(MinCutter, self).__init__()
        self.front_length = sum(l for l in lengths if l > 0)
        self.back_length = sum(l for l in lengths if l < 0)
        self.count_trimmed = count_trimmed
        self.only_trimmed = only_trimmed
    
    def __call__(self, read):
        trim_front = trim_back = True
        if self.only_trimmed:
            if read.match:
                is_front = [match.is_front for match in read.match_info]
                if not any(is_front):
                    trim_front = False
                elif all(is_front):
                    trim_back = False
            else:
                return read
        
        # TODO: distinguish between adapter trimming and other trimming.
        # For things like Methyl-Seq, we want to trim additional bases
        # after adapter trimming, but we want to count other post-adapter
        # trimmed bases toward the minimum length
        def to_trim(offset, is_front):
            if self.count_trimmed:
                trimmed = read.clipped[offset] + read.clipped[offset+2]
                if read.match:
                    trimmed += sum(i.rsize_total for i in read.match_info if is_front == i.is_front)
            elif read.match:
                trimmed = read.clipped[offset+2]
            else:
                trimmed = read.clipped[offset]
            
            if is_front:
                return max(self.front_length - trimmed, 0)
            else:
                return min(trimmed + self.back_length, 0)
        
        return self.clip(
            read,
            to_trim(0, True) if trim_front else 0,
            to_trim(1, False) if trim_back else 0
        )

class LengthTagModifier(object):
    """
    Replace "length=..." strings in read names.
    """
    def __init__(self, length_tag="length="):
        self.regex = re.compile(r"\b" + length_tag + r"[0-9]*\b")
        self.length_tag = length_tag

    def __call__(self, read):
        read = read[:]
        if read.name.find(self.length_tag) >= 0:
            read.name = self.regex.sub(self.length_tag + str(len(read.sequence)), read.name)
        return read

class SuffixRemover(object):
    """
    Remove a given suffix from read names.
    """
    def __init__(self, suffixes=[]):
        self.suffixes = suffixes

    def __call__(self, read):
        name = read.name
        for s in self.suffixes:
            if name.endswith(s):
                name = name[:-len(s)]
        read = read[:]
        read.name = name
        return read

class PrefixSuffixAdder(object):
    """
    Add a suffix and a prefix to read names
    """
    def __init__(self, prefix="", suffix=""):
        self.prefix = prefix
        self.suffix = suffix

    def __call__(self, read):
        read = read[:]
        adapter_name = 'no_adapter' if read.match is None else read.match.adapter.name
        read.name = self.prefix.replace('{name}', adapter_name) + read.name + \
            self.suffix.replace('{name}', adapter_name)
        return read

class DoubleEncoder(object):
    """
    Double-encode colorspace reads, using characters ACGTN to represent colors.
    """
    def __init__(self):
        self.double_encode_trans = str.maketrans('0123.', 'ACGTN')

    def __call__(self, read):
        read = read[:]
        read.sequence = read.sequence.translate(self.double_encode_trans)
        return read

class ZeroCapper(object):
    """
    Change negative quality values of a read to zero
    """
    def __init__(self, quality_base=33):
        qb = quality_base
        self.zero_cap_trans = str.maketrans(''.join(map(chr, range(qb))), chr(qb) * qb)

    def __call__(self, read):
        read = read[:]
        read.qualities = read.qualities.translate(self.zero_cap_trans)
        return read

class PrimerTrimmer(Trimmer):
    display_str = "Primer-trimmed"
    
    def __call__(self, read):
        """Trim primer base from colorspace reads"""
        read = self.clip(read, 1)
        read.primer = ''
        return read

class NextseqQualityTrimmer(Trimmer):
    display_str = "Quality trimmed (NextSeq)"
    
    def __init__(self, cutoff=0, base=33):
        super(NextseqQualityTrimmer, self).__init__()
        self.cutoff = cutoff
        self.base = base
    
    def __call__(self, read):
        if len(read) == 0:
            return read
        stop = nextseq_trim_index(read, self.cutoff, self.base)
        return self.subseq(read, end=stop)

class QualityTrimmer(Trimmer):
    display_str = "Quality-trimmed"
    
    def __init__(self, cutoff_front=0, cutoff_back=0, base=33):
        super(QualityTrimmer, self).__init__()
        self.cutoff_front = cutoff_front
        self.cutoff_back = cutoff_back
        self.base = base

    def __call__(self, read):
        if len(read) == 0:
            return read
        start, stop = quality_trim_index(read.qualities, self.cutoff_front, self.cutoff_back, self.base)
        return self.subseq(read, start, stop)

class NEndTrimmer(Trimmer):
    """Trims Ns from the 3' and 5' end of reads"""
    
    display_str = "End Ns trimmed"
    
    def __init__(self):
        super(NEndTrimmer, self).__init__()
        self.start_trim = re.compile(r'^N+')
        self.end_trim = re.compile(r'N+$')

    def __call__(self, read):
        if len(read) == 0:
            return read
        sequence = read.sequence
        start_cut = self.start_trim.match(sequence)
        end_cut = self.end_trim.search(sequence)
        start_cut = start_cut.end() if start_cut else 0
        end_cut = end_cut.start() if end_cut else len(read)
        return self.subseq(read, start_cut, end_cut)

class RRBSTrimmer(MinCutter):
    """
    Sequences that are adapter-trimmed are further trimmed 2 bp on the 3' end to
    remove potential methylation-biased bases from the end-repair reaction.
    """
    
    display_str = "RRBS-trimmed"
    
    def __init__(self, trim_5p=0, trim_3p=2):
        super(RRBSTrimmer, self).__init__((trim_5p, -1 * trim_3p), count_trimmed=False, only_trimmed=True)

class NonDirectionalBisulfiteTrimmer(object):
    """
    For non-directional RRBS/WGBS libraries (which implies that they were digested
    using MspI), sequences that start with either 'CAA' or 'CGA' will have 2 bp
    trimmed off the 5' end to remove potential methylation-biased bases from the
    end-repair reaction. Additionally, for RRBS reads, if CAA/CGA is not trimmed
    *and* the read has been adapter-trimmed, a minimum number of bases is trimmed
    from the 3' end.
    """
    
    display_str = "Bisulfite-trimmed (Non-directional)"
    _regex = re.compile("^C[AG]A")
    
    def __init__(self, trim_5p=2, trim_3p=2, rrbs=False):
        self._non_directional_cutter = MinCutter([trim_5p], count_trimmed=False, only_trimmed=False)
        self.rrbs = rrbs
        if rrbs:
            self._rrbs_cutter = RRBSTrimmer(trim_3p)
    
    def __call__(self, read):
        if len(read) == 0:
            return read
        cutter = None
        if self._regex.match(read.sequence):
            cutter = self._non_directional_cutter
        elif self.rrbs:
            cutter = self._rrbs_cutter
        return cutter(read) if cutter else read
    
    @property
    def modified_bases(self):
        return self._rrbs_cutter.trimmed_bases + self._non_directional_cutter.trimmed_bases

class TruSeqBisulfiteTrimmer(MinCutter):
    """EpiGnome reads are trimmed at least 6 bp on the 5' end."""
    
    display_str = "Bisulfite-trimmed (EpiGnome/TruSeq)"
    
    def __init__(self):
        super(TruSeqBisulfiteTrimmer, self).__init__((6,), count_trimmed=True, only_trimmed=False)

class SwiftBisulfiteTrimmer(ReadPairModifier):
    """
    For WGBS libraries prepared with the Swift Accel-NGS kit, 10 bp are trimmed off the end
    of read1 and the beginning of read2.
    """
    
    display_str = "Bisulfite-trimmed (Swift)"
    
    def __init__(self, trim_5p1=0, trim_3p1=10, trim_5p2=10, trim_3p2=0):
        self._read1_cutter = MinCutter((trim_5p1, -1 * trim_3p1), count_trimmed=False, only_trimmed=False)
        self._read2_cutter = MinCutter((trim_5p2, -1 * trim_3p2), count_trimmed=False, only_trimmed=False)
    
    def __call__(self, read1, read2):
        return (self._read1_cutter(read1), self._read2_cutter(read2))

    @property
    def modified_bases(self):
        return self._read1_cutter.trimmed_bases + self._read2_cutter.trimmed_bases

# TODO: InsertAdapterCutter should save the insert match, and
# MergeOverlapping should use that rather than doing another alignment

class MergeOverlapping(ReadPairModifier, ErrorCorrectorMixin):
    def __init__(self, min_overlap=0.9, error_rate=0.1, mismatch_action=None):
        ErrorCorrectorMixin.__init__(self, mismatch_action)
        self.min_overlap = int(min_overlap) if min_overlap > 1 else min_overlap
        self.error_rate = error_rate
    
    def __call__(self, read1, read2):
        l1 = len(read1.sequence)
        l2 = len(read2.sequence)
        min_overlap = self.min_overlap
        if min_overlap <= 1:
            min_overlap = max(2, round(self.min_overlap * min(l1, l2)))
        
        if l1 < min_overlap or l2 < min_overlap:
            return (read1, read2)
        
        insert_matched = read1.insert_overlap and read2.insert_overlap
        
        if insert_matched:
            # If we've already determined that there is an insert overlap
            # with a 3' overhang, we can constrain our alignment
            aflags = START_WITHIN_SEQ1 | STOP_WITHIN_SEQ2
        else:
            aflags = SEMIGLOBAL
        # align read1 to read2 reverse-complement to be compatible with InsertAligner
        read2_rc = reverse_complement(read2.sequence)
        aligner = Aligner(read2_rc, self.error_rate, aflags)
        alignment = aligner.locate(read1.sequence)
        
        if alignment:
            r2_start, r2_stop, r1_start, r1_stop, matches, errors = alignment
            if matches >= min_overlap:
                # Only correct errors if we haven't already done correction in the InsertAligner
                if (self.mismatch_action and errors > 0 and not insert_matched and
                        read1.corrected == 0 and read2.corrected == 0):
                    self.correct_errors(read1, read2, alignment)
                
                if r2_start == 0 and r2_stop == l2:
                    # r2 is fully contained in r1
                    pass
                elif r1_start == 0 and r1_stop == l1:
                    # r1 is fully contained in r2
                    read1.sequence = read2_rc
                    read1.qualities = "".join(reversed(read2.qualities))
                elif r1_start > 0:
                    read1.sequence += read2_rc[r2_stop:]
                    if read1.qualities and read2.qualities:
                        read1.qualities += "".join(reversed(read2.qualities))[r2_stop:]
                elif r2_start > 0:
                    read1.sequence = read2_rc + read1.sequence[r1_stop:]
                    if read1.qualities and read2.qualities:
                        read1.qualities = "".join(reversed(read2.qualities)) + read1.qualities[r1_stop:]
                else:
                    raise Exception(
                        "Invalid alignment while trying to merge read {}: {}".format(
                            read1.name, ",".join(str(i) for i in alignment)))
                
                read1.merged = True
                read2 = None
                
        return (read1, read2)

class Modifiers(object):
    def __init__(self, paired):
        self.modifiers = []
        self.modifier_indexes = {}
        self.paired = paired
    
    def add_modifier(self, mod_class, read=1|2, **kwargs):
        if issubclass(mod_class, ReadPairModifier):
            assert self.paired == "both" and read == 1|2
            mods = mod_class(**kwargs)
        else:
            mods = [None, None]
            if read & 1 > 0:
                mods[0] = mod_class(**kwargs)
            if read & 2 > 0 and self.paired == "both":
                mods[1] = mod_class(**kwargs)
            if all(m is None for m in mods):
                return None
        return self._add_modifiers(mod_class, mods)
    
    def add_modifier_pair(self, mod_class, read1_args=None, read2_args=None):
        mods = [None, None]
        if read1_args is not None:
            mods[0] = mod_class(**read1_args)
        if read2_args is not None and self.paired == "both":
            mods[1] = mod_class(**read2_args)
        if all(m is None for m in mods):
            return None
        return self._add_modifiers(mod_class, mods)
    
    def _add_modifiers(self, mod_class, mods):
        i = len(self.modifiers)
        self.modifiers.append(mods)
        if mod_class in self.modifier_indexes:
            self.modifier_indexes[mod_class].append(i)
        else:
            self.modifier_indexes[mod_class] = [i]
        return i
    
    def get_modifiers(self, mod_class=None, read=None):
        if mod_class is None:
            mods = copy.copy(self.modifiers)
        elif mod_class in self.modifier_indexes:
            mods = [self.modifiers[i] for i in self.modifier_indexes[mod_class]]
        else:
            mods = []
        
        if not (mods and read):
            return mods
        
        read_mods = []
        for m in mods:
            if isinstance(m, ReadPairModifier):
                read_mods.append(m)
            elif m[read-1] is not None:
                read_mods.append(m[read-1])
        return read_mods
    
    def has_modifier(self, mod_class):
        return mod_class in self.modifier_indexes
    
    def get_adapters(self):
        adapters = [[], []]
        if self.has_modifier(AdapterCutter):
            m1, m2 = self.get_modifiers(AdapterCutter)[0]
            if m1:
                adapters[0] = m1.adapters
            if m2:
                adapters[1] = m2.adapters
        elif self.has_modifier(InsertAdapterCutter):
            m = self.get_modifiers(InsertAdapterCutter)[0]
            adapters[0] = [m.adapter1]
            adapters[1] = [m.adapter2]
        return adapters
    
    def get_trimmer_classes(self):
        return [
            klass for klass in self.modifier_indexes.keys()
            if issubclass(klass, Trimmer)
        ]
    
    def modify(self, record):
        bp = [0,0]
        if self.paired:
            read1, read2 = record
            bp[0] = len(read1.sequence)
            bp[1] = len(read2.sequence)
            for mods in self.modifiers:
                if isinstance(mods, ReadPairModifier):
                    read1, read2 = mods(read1, read2)
                else:
                    if mods[0] is not None:
                        read1 = mods[0](read1)
                    if mods[1] is not None:
                        read2 = mods[1](read2)
            reads = [read1, read2]
        else:
            read = record
            bp[0] = len(read.sequence)
            for mods in self.modifiers:
                read = mods[0](read)
            reads = [read]
        return (reads, bp)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# kate: word-wrap off; remove-trailing-spaces all;

from argparse import ArgumentParser, HelpFormatter
from collections import OrderedDict
import copy
import logging
import operator
import os
import platform
import re
import sys
import time
import textwrap

# Print a helpful error message if the extension modules cannot be imported.
from atropos import *
check_importability()

from atropos import __version__
import atropos.commands
from atropos.util import MAGNITUDE
from atropos.xopen import STDOUT, STDERR, resolve_path, check_path, check_writeable

# Extensions to argparse

class ParagraphHelpFormatter(HelpFormatter):
    def _fill_text(self, text, width, indent):
        text = re.sub('[ \t]{2,}', ' ', text)
        paragraphs = [
            textwrap.fill(p, width, initial_indent=indent, subsequent_indent=indent)
            for p in re.split("\n\n", text)
        ]
        return "\n\n".join(paragraphs)

class TypeWithArgs(object):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __call__(self, string):
        return self._do_call(string, *self.args, **self.kwargs) or string

class CompositeType(object):
    def __init__(self, *types):
        self.types = types

    def __call__(self, string):
        result = string
        for t in self.types:
            result = t(result)
        return result

class ComparisonValidator(TypeWithArgs):
    def _do_call(self, lhs, rhs, oper, expected=True):
        assert oper(lhs, rhs) == expected, "{}({}, {}) != {}".format(oper, lhs, rhs, expected)

class char_list(object):
    def __init__(self, choices):
        self.choices = set(choices)
    
    def __call__(self, string):
        l = list(string)
        assert all(c in self.choices for c in l)
        return l

class delimited(TypeWithArgs):
    """Splits a string argument using a delimiter."""
    def _do_call(self, string, delim=",", data_type=None, choices=None, min_len=None, max_len=None):
        if isinstance(string, str):
            vals = string.split(delim) if delim else (string,)
        else:
            vals = string
        
        if vals[0] == "*" and choices is not None:
            vals = choices
        
        if data_type:
            vals = [data_type(v) for v in vals]
        
        if min_len and len(vals) < min_len:
            raise ArgumentError(self, "there must be at least {} values".format(min_len))
        
        if max_len and len(vals) > max_len:
            raise ArgumentError(self, "there can be at most {} values".format(max_len))
        
        return vals

ACCESS = dict(r=os.R_OK, rU=os.R_OK, rb=os.R_OK, w=os.W_OK, wb=os.W_OK, x=os.X_OK)

class accessible_path(TypeWithArgs):
    """Test that a path is accessible"""
    def _do_call(self, path, type_, mode):
        if type_ == 'f' and path in (STDOUT, STDERR):
            return path
        if 'w' in mode:
            return check_writeable(path, type_)
        else:
            return check_path(path, type_, ACCESS[mode])

def existing_path(path):
    """Test that a path exists."""
    if path == STDOUT:
        return path
    return resolve_path(path)

readable_file = CompositeType(existing_path, accessible_path('f', 'r'))
"""Test that a file exists and is readable."""

writeable_file = accessible_path('f', 'w')
"""
Test that a file 1) exists and is writelable, or 2) does not exist but
is in a writeable directory.
"""

class _readwriteable_file(object):
    def __init__(self):
        self.r = accessible_path('f', 'r')
        self.w = accessible_path('f', 'w')
    
    def __call__(self, string):
        path = string
        if os.path.exists(path):
            path = self.r(path)
        path = self.w(path)
        return path
readwriteable_file = _readwriteable_file()
"""Test that a file is both readable and writeable."""

import urllib
def readable_url(url):
    parsed = urllib.parse.urlparse(url)
    scheme = parsed.scheme or 'file'
    if scheme == 'file':
        filename = readable_file(parsed.path)
        return 'file:' + filename
    else:
        return url

str_list = delimited(data_type=str)
"""Comma-delimited list of strings."""

int_or_str_re = re.compile("([\d\.]+)([KkMmGg]?)")
def int_or_str(x):
    """Similar to int(), but accepts K, M, and G abbreviations"""
    if x is None or isinstance(x, int):
        return x
    elif isinstance(x, str):
        match = int_or_str_re.match(x.upper())
        num, mult = match.groups()
        if mult:
            return int(float(num) * MAGNITUDE[mult])
        else:
            return int(num)
    else:
        raise Exception("Unsupported type {}".format(x))

def positive(type_=int, inclusive=False):
    """Test that a number is greater than (or equal to, if ``inclusive=True``) zero."""
    return ge(0, type_) if inclusive else gt(0, type_)

def gt(x, type_=int):
    """Test that a number is greater than another number."""
    return CompositeType(type_, ComparisonValidator(x, operator.gt))

def ge(x, type_=int):
    """Test that a number is greater than another number."""
    return CompositeType(type_, ComparisonValidator(x, operator.ge))

def between(min_val=None, max_val=None, type_=int):
    return CompositeType(
        type_,
        ComparisonValidator(min_val, operator.ge),
        ComparisonValidator(max_val, operator.le))

probability = between(0, 1, float)
"""A float between 0-1 (inclusive)."""

# Commands

COMMANDS = OrderedDict()

class Command(object):
    def __init__(self, args):
        self.orig_args = copy.copy(args)
        self.create_parser()
        self.add_common_options()
        self.add_command_options()
        self.options = self.parser.parse_args(args)
        self.setup_logging()
        self.validate_common_options()
        self.validate_command_options()
    
    def create_parser(self):
        self.parser = ArgumentParser(
            usage=self.usage,
            description=self.description.format(version=__version__),
            formatter_class=ParagraphHelpFormatter)
    
    def add_common_options(self):
        # Add common options
        self.parser.set_defaults(
            paired=False,
            default_outfile=STDOUT,
            batch_size=1000)
        self.parser.add_argument(
            "--debug",
            action='store_true', default=False,
            help="Print debugging information. (no)")
        self.parser.add_argument(
            "--progress",
            choices=('bar', 'msg'), default=None,
            help="Show progress. bar = show progress bar; msg = show a status message. (no)")
        self.parser.add_argument(
            "--quiet",
            action='store_true', default=False,
            help="Print only error messages. (no)")
        self.parser.add_argument(
            "--log-level",
            choices=('DEBUG', 'INFO', 'WARN', 'ERROR'), default=None,
            help="Logging level. (ERROR when --quiet else INFO)")
        self.parser.add_argument(
            "--log-file",
            type=writeable_file, default=None, metavar="FILE",
            help="File to write logging info. (stdout)")
        
        group = self.parser.add_argument_group("Input")
        group.add_argument(
            "-pe1",
            "--input1",
            type=readable_file, default=None, metavar="FILE1",
            help="The first (and possibly only) input file.")
        group.add_argument(
            "-pe2",
            "--input2",
            type=readable_file, default=None, metavar="FILE2",
            help="The second input file.")
        group.add_argument(
            "-l",
            "--interleaved-input",
            type=readable_file, default=None, metavar="FILE",
            help="Interleaved input file.")
        group.add_argument(
            "-se",
            "--single-input",
            type=readable_file, default=None, metavar="FILE",
            help="A single-end read file.")
        group.add_argument(
            "--single-input-read",
            type=int, choices=(1, 2), default=None,
            help="When treating an interleaved FASTQ or paired-end SAM/BAM "
                "file as single-end, this option specifies which of the two "
                "reads to process. (both reads used)")
        group.add_argument(
            "-sq",
            "--single-quals",
            type=readable_file, default=None, metavar="FILE",
            help="A single-end qual file.")
        group.add_argument(
            "-f",
            "--format",
            choices=('fasta','fastq','sra-fastq','sam','bam'), default=None,
            help="Input file format. Ignored when reading csfasta/qual files. (auto-detect from "
                 "file name extension)")
        group.add_argument(
            "-c",
            "--colorspace",
            action='store_true', default=False,
            help="Enable colorspace mode: Also trim the color that is adjacent to "
                 "the found adapter. (no)")
        group.add_argument(
            "--max-reads",
            type=int_or_str, default=None, metavar="N",
            help="Maximum number of reads/pairs to process (no max)")
        group.add_argument(
            "--subsample",
            type=probability, default=None, metavar="PROB",
            help="Subsample a fraction of reads. (no)")
        group.add_argument(
            "--batch-size",
            type=int_or_str, metavar="SIZE",
            help="Number of records to process in each batch. (5000)")
    
    def add_command_options(self):
        """
        Add command-specific options. At the very least,
        "-o, --output" is required.
        """
        raise NotImplemented()
    
    def setup_logging(self):
        # Setup logging only if there are not already any handlers (can happen when
        # this function is being called externally such as from unit tests)
        if not logging.root.handlers:
            level = self.options.log_level or ("ERROR" if self.options.quiet else "INFO")
            level = getattr(logging, level)
            if self.options.log_file is None:
                # Due to backwards compatibility, log messages are sent to standard output
                # instead of standard error if the -o option is used.
                handler = logging.StreamHandler(
                    sys.stdout if self.options.output not in (None, STDOUT, STDERR) else sys.stderr)
            else:
                handler = logging.FileHandler(options.log_file)
            handler.setFormatter(logging.Formatter('%(asctime)s %(levelname)s: %(message)s'))
            handler.setLevel(level)
            logging.getLogger().setLevel(level)
            logging.getLogger().addHandler(handler)
        
        logger = logging.getLogger()
        logger.info("This is Atropos %s with Python %s", __version__, platform.python_version())
        logger.info("Command line parameters: %s %s", self.name, " ".join(self.orig_args))
    
    def validate_common_options(self):
        options = self.options
        parser = self.parser
        
        # Find out which 'mode' we need to use.
        if options.single_input:
            if options.input1 or options.input2 or options.interleaved_input:
                parser.error("Cannot use -se together with -pe1, -pe2, or -l")
            options.paired = False
            options.input1 = options.single_input
            options.input2 = options.single_quals
        elif options.interleaved_input and options.single_input_read:
            options.input1 = options.interleaved_input
            options.paired = False
        else:
            if not options.interleaved_input and (not options.input1 or not options.input2):
                parser.error("Both '-pe1' and '-pe2' are required for paired-end trimming. If this is an "
                    "interleaved file, use '-l' instead.")
            options.paired = True
        
        if options.quiet:
            options.progress = None
        elif options.progress and options.output == STDERR:
            parser.warn("Progress bar may corrupt output written to STDERR")
    
    def validate_command_options(self):
        pass
    
    def execute(self):
        cmd = getattr(atropos.commands, self.name)
        cmd(self.options, self.parser)

class TrimCommand(Command):
    """trim sequencing reads."""
    name = "trim"
    usage = """
atropos trim -a ADAPTER [options] [-o output.fastq] -se input.fastq
atropos trim -a ADAPT1 -A ADAPT2 [options] -o out1.fastq -p out2.fastq -pe1 in1.fastq -pe2 in2.fastq
"""
    description = """
Atropos version {version}

Replace "ADAPTER" with the actual sequence of your 3' adapter. IUPAC wildcard
characters are supported. The reverse complement is *not* automatically
searched. All reads from input.fastq will be written to output.fastq with the
adapter sequence removed. Adapter matching is error-tolerant. Multiple adapter
sequences can be given (use further -a options), but only the best-matching
adapter will be removed.

Input may also be in FASTA format. Compressed input and output is supported and
auto-detected from the file name (.gz, .xz, .bz2). Use the file name '-' for
standard input/output. Without the -o option, output is sent to standard output.
"""
    
    def add_command_options(self):
        parser = self.parser
        parser.set_defaults(
            zero_cap=None,
            action='trim',
            batch_size=5000,
            known_adapter=None)
        
        group = parser.add_argument_group("Finding adapters",
            description="Parameters -a, -g, -b specify adapters to be removed from "
                "each read (or from the first read in a pair if data is paired). "
                "If specified multiple times, only the best matching adapter is "
                "trimmed (but see the --times option). When the special notation "
                "'file:FILE' is used, adapter sequences are read from the given "
                "FASTA file. When the --adapter-file option is used, adapters can "
                "be specified by name rather than sequence.")
        group.add_argument(
            "-a",
            "--adapter",
            action="append", default=[], metavar="ADAPTER", dest="adapters",
            help="Sequence of an adapter ligated to the 3' end (paired data: of the "
                "first read). The adapter and subsequent bases are trimmed. If a "
                "'$' character is appended ('anchoring'), the adapter is only "
                "found if it is a suffix of the read. (none)")
        group.add_argument(
            "-g",
            "--front",
            action="append", default=[], metavar="ADAPTER",
            help="Sequence of an adapter ligated to the 5' end (paired data: of the "
                "first read). The adapter and any preceding bases are trimmed. "
                "Partial matches at the 5' end are allowed. If a '^' character is "
                "prepended ('anchoring'), the adapter is only found if it is a "
                "prefix of the read. (none)")
        group.add_argument(
            "-b",
            "--anywhere",
            action="append", default=[], metavar="ADAPTER",
            help="Sequence of an adapter that may be ligated to the 5' or 3' end "
                "(paired data: of the first read). Both types of matches as "
                "described under -a und -g are allowed. If the first base of the "
                "read is part of the match, the behavior is as with -g, otherwise "
                "as with -a. This option is mostly for rescuing failed library "
                "preparations - do not use if you know which end your adapter was "
                "ligated to! (none)")
        group.add_argument(
            "-F",
            "--known-adapters-file",
            type=readable_file, action="append", default=None,
            help="Path or URL of a FASTA file containing adapter sequences.")
        group.add_argument(
            "--no-default-adapters",
            action="store_false", dest="default_adapters", default=True,
            help="Don't fetch the default adapter list (which is currently stored as a) "
                "GitHub gist).")
        group.add_argument(
            "--adapter-cache-file",
            type=readwriteable_file, default='.adapters',
            help="File where adapter sequences will be cached, unless "
                "--no-cache-adapters is set.")
        group.add_argument(
            "--no-cache-adapters",
            action="store_false", dest="cache_adapters", default=True,
            help="Don't cache adapters list as '.adapters' in the working directory.")
        group.add_argument(
            "--no-trim",
            action='store_const', dest='action', const=None,
            help="Match and redirect reads to output/untrimmed-output as usual, "
                "but do not remove adapters. (no)")
        group.add_argument(
            "--mask-adapter",
            action='store_const', dest='action', const='mask',
            help="Mask adapters with 'N' characters instead of trimming them. (no)")
        
        ## Arguments specific to the choice of aligner
        
        group.add_argument("--aligner", choices=('adapter', 'insert'), default='adapter',
            help="Which alignment algorithm to use for identifying adapters. Currently, "
                 "you can choose between the semi-global alignment strategy used in Cutdapt "
                 "('adapter') or the more accurate insert-based alignment algorithm ('insert'). "
                 "Note that insert-based alignment can only be used with paired-end reads "
                 "containing 3' adapters. New algorithms are being implemented and the default "
                 "is likely to change. (adapter)")
        
        # TODO: all the different matching options are pretty confusing. Either explain their
        # usage better in the docs or find a way to simplify the choices.
        
        # Arguments for adapter match
        group.add_argument(
            "-e",
            "--error-rate",
            type=probability, default=None,
            help="Maximum allowed error rate for adapter match (no. of errors divided by the length "
                "of the matching region). (0.1)")
        group.add_argument(
            "--indel-cost",
            type=positive(int, True), default=None, metavar="COST",
            help="Integer cost of insertions and deletions during adapter match. Substitutions always "
                 "have a cost of 1. (1)")
        group.add_argument(
            "--no-indels",
            action='store_false', dest='indels', default=True,
            help="Allow only mismatches in alignments. (allow both mismatches and indels)")
        group.add_argument(
            "-n",
            "--times",
            type=positive(int, False), default=1, metavar="COUNT",
            help="Remove up to COUNT adapters from each read. (1)")
        group.add_argument(
            "--match-read-wildcards",
            action="store_true", default=False,
            help="Interpret IUPAC wildcards in reads. (no)")
        group.add_argument(
            "-N",
            "--no-match-adapter-wildcards",
            action="store_false", dest='match_adapter_wildcards', default=True,
            help="Do not interpret IUPAC wildcards in adapters. (no)")
        group.add_argument(
            "-O",
            "--overlap",
            type=positive(int, False), default=None, metavar="MINLENGTH",
            help="If the overlap between the read and the adapter is shorter than "
                "MINLENGTH, the read is not modified. Reduces the no. of bases "
                "trimmed due to random adapter matches. (3)")
        group.add_argument(
            "--adapter-max-rmp",
            type=probability, default=None, metavar="PROB",
            help="If no minimum overlap (-O) is specified, then adapters are only matched "
                 "when the probabilty of observing k out of n matching bases is <= PROB. (1E-6)")
        
        # Arguments for insert match
        group.add_argument(
            "--insert-max-rmp",
            type=probability, default=1E-6, metavar="PROB",
            help="Overlapping inserts only match when the probablity of observing k of n "
                 "matching bases is <= PROB. (1E-6)")
        group.add_argument(
            "--insert-match-error-rate",
            type=probability, default=None,
            help="Maximum allowed error rate for insert match (no. of errors divided by the length "
                "of the matching region). (0.2)")
        group.add_argument(
            "--insert-match-adapter-error-rate",
            type=probability, default=None,
            help="Maximum allowed error rate for matching adapters after successful insert match "
                 "(no. of errors divided by the length of the matching region). (0.2)")
        
        # Arguments for merging and error correction
        # TODO: add RMP parameter for MergeOverlap
        group.add_argument(
            "-R",
            "--merge-overlapping",
            action="store_true", default=False,
            help="Merge read pairs that overlap into a single sequence. This is experimental. (no)")
        group.add_argument(
            "--merge-min-overlap",
            type=positive(float, True), default=0.9,
            help="The minimum overlap between reads required for merging. If this number is (0,1.0], "
                 "it specifies the minimum length as the fraction of the length of the *shorter* read "
                 "in the pair; otherwise it specifies the minimum number of overlapping base pairs ("
                 "with an absolute minimum of 2 bp). (0.9)")
        group.add_argument(
            "--merge-error-rate",
            type=probability, default=None,
            help="The maximum error rate for merging. (0.2)")
        group.add_argument(
            "--correct-mismatches",
            choices=("liberal", "conservative", "N"), default=None,
            help="How to handle mismatches while aligning/merging. 'Liberal' and 'conservative' error "
                 "correction both involve setting the base to the one with the best quality. They differ "
                 "only when the qualities are equal -- liberal means set it to the base from the read with "
                 "the overall best median base quality, while conservative means to leave it unchanged. "
                 "'N' means to set the base to N. If exactly one base is ambiguous, the non-ambiguous base "
                 "is always used. (no error correction)")
        
        group = parser.add_argument_group("Additional read modifications")
        group.add_argument(
            "--op-order",
            type=char_list(choices=('A','C','G','Q')), default="CGQA",
            help="The order in which trimming operations are be applied. This is a string of "
                 "1-4 of the following characters: A = adapter trimming; C = cutting "
                 "(unconditional); G = NextSeq trimming; Q = quality trimming. The default is "
                 "'CGQA' to maintain compatibility with Cutadapt; however, this is likely to "
                 "change to 'GACQ' in the near future.")
        group.add_argument(
            "-u",
            "--cut",
            type=int, action='append', default=[], metavar="LENGTH",
            help="Remove bases from each read (first read only if paired). "
                "If LENGTH is positive, remove bases from the beginning. "
                "If LENGTH is negative, remove bases from the end. "
                "Can be used twice if LENGTHs have different signs. (no)")
        group.add_argument(
            "-q",
            "--quality-cutoff",
            type=delimited(data_type=positive(int, True), min_len=1, max_len=2),
            default=None, metavar="[5'CUTOFF,]3'CUTOFF",
            help="Trim low-quality bases from 5' and/or 3' ends of each read before "
                "adapter removal. Applied to both reads if data is paired. If one "
                "value is given, only the 3' end is trimmed. If two "
                "comma-separated cutoffs are given, the 5' end is trimmed with "
                "the first cutoff, the 3' end with the second. (no)")
        group.add_argument(
            "-i",
            "--cut-min",
            type=int, action='append', default=[], metavar="LENGTH",
            help="Similar to -u, except that cutting is done AFTER adapter trimming, and only "
                "if a minimum of LENGTH bases was not already removed. (no)")
        group.add_argument(
            "--nextseq-trim",
            type=positive(), default=None, metavar="3'CUTOFF",
            help="NextSeq-specific quality trimming (each read). Trims also dark "
                "cycles appearing as high-quality G bases (EXPERIMENTAL). (no)")
        group.add_argument(
            "--quality-base",
            type=positive(), default=33,
            help="Assume that quality values in FASTQ are encoded as ascii(quality "
                "+ QUALITY_BASE). This needs to be set to 64 for some old Illumina "
                "FASTQ files. (33)")
        group.add_argument(
            "--trim-n",
             action='store_true', default=False,
            help="Trim N's on ends of reads. (no)")
        group.add_argument(
            "-x",
            "--prefix",
            default='',
            help="Add this prefix to read names. Use {name} to insert the name of "
                 "the matching adapter. (no)")
        group.add_argument(
            "-y",
            "--suffix",
            default='',
            help="Add this suffix to read names; can also include {name}. (no)")
        group.add_argument(
            "--strip-suffix",
            action='append', default=[],
            help="Remove this suffix from read names if present. Can be given "
                 "multiple times. (no)")
        group.add_argument(
            "--length-tag",
            metavar="TAG",
            help="Search for TAG followed by a decimal number in the description "
                "field of the read. Replace the decimal number with the correct "
                "length of the trimmed read. For example, use --length-tag 'length=' "
                "to correct fields like 'length=123'. (no)")
        
        group = parser.add_argument_group("Filtering of processed reads")
        group.add_argument(
            "--discard-trimmed", "--discard",
            action='store_true', default=False,
            help="Discard reads that contain an adapter. Also use -O to avoid "
                "discarding too many randomly matching reads! (no)")
        group.add_argument(
            "--discard-untrimmed", "--trimmed-only",
            action='store_true', default=False,
            help="Discard reads that do not contain the adapter. (no)")
        group.add_argument(
            "-m",
            "--minimum-length",
            type=positive(int, True), default=None, metavar="LENGTH",
            help="Discard trimmed reads that are shorter than LENGTH. Reads that "
                "are too short even before adapter removal are also discarded. In "
                "colorspace, an initial primer is not counted. (0)")
        group.add_argument(
            "-M",
            "--maximum-length",
            type=positive(int, True), default=sys.maxsize, metavar="LENGTH",
            help="Discard trimmed reads that are longer than LENGTH. "
                "Reads that are too long even before adapter removal "
                "are also discarded. In colorspace, an initial primer "
                "is not counted. (no limit)")
        group.add_argument(
            "--max-n",
            type=positive(float, True), default=None, metavar="COUNT",
            help="Discard reads with too many N bases. If COUNT is an integer, it "
                "is treated as the absolute number of N bases. If it is between 0 "
                "and 1, it is treated as the proportion of N's allowed in a read. "
                "(no)")
        
        group = parser.add_argument_group("Output")
        group.add_argument(
            "-o",
            "--output",
            type=writeable_file, metavar="FILE",
            help="Write trimmed reads to FILE. FASTQ or FASTA format is chosen "
                "depending on input. The summary report is sent to standard output. "
                "Use '{name}' in FILE to demultiplex reads into multiple "
                "files. (write to standard output)")
        group.add_argument(
            "--info-file",
            type=writeable_file, metavar="FILE",
            help="Write information about each read and its adapter matches into FILE. "
                 "See the documentation for the file format. (no)")
        group.add_argument(
            "-r",
            "--rest-file",
            type=writeable_file, metavar="FILE",
            help="When the adapter matches in the middle of a read, write the "
                "rest (after the adapter) into FILE. (no)")
        group.add_argument(
            "--wildcard-file",
            type=writeable_file, metavar="FILE",
            help="When the adapter has N bases (wildcards), write adapter bases "
                "matching wildcard positions to FILE. When there are indels in the "
                "alignment, this will often not be accurate. (no)")
        group.add_argument(
            "--too-short-output",
            type=writeable_file, metavar="FILE",
            help="Write reads that are too short (according to length specified by "
                "-m) to FILE. (no - too short reads are discarded)")
        group.add_argument(
            "--too-long-output",
            type=writeable_file, metavar="FILE",
            help="Write reads that are too long (according to length specified by "
                "-M) to FILE. (no - too long reads are discarded)")
        group.add_argument(
            "--untrimmed-output",
            type=writeable_file, default=None, metavar="FILE",
            help="Write reads that do not contain the adapter to FILE. (no - "
                "untrimmed reads are written to default output)")
        group.add_argument(
            "--merged-output",
            type=writeable_file, default=None, metavar="FILE",
            help="Write reads that have been merged to this file. (merged reads "
                "are discarded)")
        group.add_argument(
            "--report-file",
            type=writeable_file, default=None, metavar="FILE",
            help="Write report to file rather than stdout/stderr. (no)")
        
        group = parser.add_argument_group("Colorspace options")
        group.add_argument(
            "-d",
            "--double-encode",
            action='store_true', default=False,
            help="Double-encode colors (map 0,1,2,3,4 to A,C,G,T,N). (no)")
        group.add_argument(
            "-t",
            "--trim-primer",
            action='store_true', default=False,
            help="Trim primer base and the first color (which is the transition "
                "to the first nucleotide). (no)")
        group.add_argument(
            "--strip-f3",
            action='store_true', default=False,
            help="Strip the _F3 suffix of read names. (no)")
        group.add_argument(
            "--maq", "--bwa",
            action='store_true', default=False,
            help="MAQ- and BWA-compatible colorspace output. This enables -c, -d, "
                "-t, --strip-f3 and -y '/1'. (no)")
        group.add_argument(
            "--no-zero-cap",
            dest='zero_cap', action='store_false',
            help="Do not change negative quality values to zero in colorspace "
                "data. By default, they are since many tools have problems with "
                "negative qualities. (no)")
        group.add_argument(
            "-z",
            "--zero-cap",
            action='store_true',
            help="Change negative quality values to zero. This is enabled "
                "by default when -c/--colorspace is also enabled. Use the above option "
                "to disable it. (no)")
        
        group = parser.add_argument_group("Paired-end options", description="The "
            "-A/-G/-B/-U options work like their -a/-b/-g/-u counterparts, but "
            "are applied to the second read in each pair.")
        group.add_argument(
            "-A",
            action='append', dest='adapters2', default=[], metavar='ADAPTER',
            help="3' adapter to be removed from second read in a pair. (no)")
        group.add_argument(
            "-G",
            action='append', dest='front2', default=[], metavar='ADAPTER',
            help="5' adapter to be removed from second read in a pair. (no)")
        group.add_argument(
            "-B",
            action='append', dest='anywhere2',  default=[], metavar='ADAPTER',
            help="5'/3 adapter to be removed from second read in a pair. (no)")
        group.add_argument(
            "-U",
            type=int, action='append', dest='cut2', default=[], metavar="LENGTH",
            help="Remove LENGTH bases from second read in a pair (see --cut). (no)")
        group.add_argument(
            "-I",
            "--cut-min2",
            type=int, action='append', default=[], metavar="LENGTH",
            help="Similar to -U, except that cutting is done AFTER adapter trimming, "
                 "and only if a minimum of LENGTH bases was not already removed "
                 "(see --cut-min). (no)")
        group.add_argument(
            "-p",
            "--paired-output",
            type=writeable_file, metavar="FILE",
            help="Write second read in a pair to FILE. (no)")
        group.add_argument(
            "-L",
            "--interleaved-output",
            type=writeable_file, metavar="FILE",
            help="Write output to interleaved file.")
        # Setting the default for pair_filter to None allows us to find out whether
        # the option was used at all.
        group.add_argument(
            "--pair-filter",
            choices=("any", "both"), default=None, metavar='(any|both)',
            help="Which of the reads in a paired-end read have to match the "
                "filtering criterion in order for it to be filtered. (any)")
        group.add_argument(
            "--untrimmed-paired-output",
            type=writeable_file, default=None, metavar="FILE",
            help="Write second read in a pair to this FILE when no adapter "
                "was found in the first read. Use this option together with "
                "--untrimmed-output when trimming paired-end reads. (no - output "
                "to same file as trimmed reads)")
        group.add_argument(
            "--too-short-paired-output",
            type=writeable_file, default=None, metavar="FILE",
            help="Write second read in a pair to this file if pair is too short. "
                "Use together with --too-short-output. (no - too short reads are "
                "discarded)")
        group.add_argument(
            "--too-long-paired-output",
            type=writeable_file, default=None, metavar="FILE",
            help="Write second read in a pair to this file if pair is too long. "
                "Use together with --too-long-output. (no - too long reads are "
                "discarded)")
        
        group = parser.add_argument_group("Method-specific options")
        group = group.add_mutually_exclusive_group()
        group.add_argument(
            "--bisulfite",
            default=False, metavar="METHOD",
            help="Set default option values for bisulfite-treated data. The argument specifies the "
                 "type of bisulfite library (rrbs, non-directional, non-directional-rrbs, truseq, "
                 "epignome, or swift) or custom parameters for trimming: '<read1>[;<read2>]' where "
                 "trimming parameters for each read are: '<5' cut>,<3' cut>,<include trimmed>,<only trimmed>' "
                 "where 'include trimmed' is 1 or 0 for whether or not the bases already trimmed "
                 "during/prior to adapter trimming should be counted towards the total bases to be "
                 "cut and 'only trimmed' is 1 or 0 for whether or not only trimmed reads should be "
                 "further cut. (no)")
        group.add_argument(
            "--mirna",
            action="store_true", default=False,
            help="Set default option values for miRNA data. (no)")
        
        group = parser.add_argument_group("Parallel (multi-core) options")
        group.add_argument(
            "-T",
            "--threads",
            type=positive(int, True), default=None, metavar="THREADS",
            help="Number of threads to use for read trimming. Set to 0 to use "
                 "max available threads. (Do not use multithreading)")
        group.add_argument(
            "--no-writer-process",
            action="store_false", dest="writer_process", default=True,
            help="Do not use a writer process; instead, each worker thread writes its "
                 "own output to a file with a '.N' suffix. (no)")
        group.add_argument(
            "--preserve-order",
            action="store_true", default=False,
            help="Preserve order of reads in input files (ignored if "
                 "--no-writer-process is set). (no)")
        group.add_argument(
            "--process-timeout",
            type=positive(int, True), default=60, metavar="SECONDS",
            help="Number of seconds process should wait before escalating messages "
                 "to ERROR level. (60)")
        group.add_argument(
            "--read-queue-size",
            type=int_or_str, default=None, metavar="SIZE",
            help="Size of queue for batches of reads to be processed. (THREADS * 100)")
        group.add_argument(
            "--result-queue-size",
            type=int_or_str, default=None, metavar="SIZE",
            help="Size of queue for batches of results to be written. (THREADS * 100)")
        group.add_argument(
            "--compression",
            choices=("worker", "writer"), default=None,
            help="Where data compression should be performed. Defaults to 'writer' if "
                 "system-level compression can be used and (1 < threads < 8), otherwise "
                 "defaults to 'worker'.")
    
    def validate_command_options(self):
        options = self.options
        parser = self.parser
        paired = options.paired
        
        if not paired:
            if not options.output:
                parser.error("An output file is required")
            if options.untrimmed_paired_output:
                parser.error("Option --untrimmed-paired-output can only be used when "
                    "trimming paired-end reads (with option -p).")
        else:
            if not options.interleaved_output:
                if not options.output:
                    parser.error("When you use -p or --paired-output, you must also "
                        "use the -o option.")
                if not options.paired_output:
                    parser.error("When paired-end trimming is enabled via -A/-G/-B/-U, "
                        "a second output file needs to be specified via -p (--paired-output).")
                if bool(options.untrimmed_output) != bool(options.untrimmed_paired_output):
                    parser.error("When trimming paired-end reads, you must use either none "
                        "or both of the --untrimmed-output/--untrimmed-paired-output options.")
                if options.too_short_output and not options.too_short_paired_output:
                    parser.error("When using --too-short-output with paired-end "
                        "reads, you also need to use --too-short-paired-output")
                if options.too_long_output and not options.too_long_paired_output:
                    parser.error("When using --too-long-output with paired-end "
                        "reads, you also need to use --too-long-paired-output")
            
            # Any of these options switch off legacy mode
            if (options.adapters2 or options.front2 or options.anywhere2 or options.cut2 or
                options.cut_min2 or options.interleaved_input or options.pair_filter or
                options.too_short_paired_output or options.too_long_paired_output):
                # Full paired-end trimming when both -p and -A/-G/-B/-U given
                # Read modifications (such as quality trimming) are applied also to second read.
                paired = 'both'
            else:
                # Modify first read only, keep second in sync (-p given, but not -A/-G/-B/-U).
                # This exists for backwards compatibility ('legacy mode').
                paired = 'first'
            
            options.paired = paired
        
        # If the user specifies a max rmp, that is used for determining the
        # minimum overlap and -O is set to 1, otherwise -O is set to the old
        # default of 3.
        # TODO: This is pretty confusing logic - need to simplify
        
        if options.aligner == 'adapter':
            if options.indels and options.indel_cost is None:
                options.indel_cost = 1
            if options.overlap is None:
                if options.adapter_max_rmp is None:
                    options.overlap = 3
                else:
                    options.overlap = 1
        elif options.aligner == 'insert':
            if paired != 'both':
                parser.error("Insert aligner only works with paired-end reads")
                # TODO: should also be checking that there is exactly one 3' adapter for each read
                # TODO: have the aligner tell us whether it can be used based on options?
            if options.indels and options.indel_cost is None:
                options.indel_cost = 3
            if options.overlap is None:
                options.overlap = 1
                if options.adapter_max_rmp is None:
                    options.adapter_max_rmp = 1E-6
            if options.insert_match_error_rate is None:
                options.insert_match_error_rate = options.error_rate or 0.2
            if options.insert_match_adapter_error_rate is None:
                options.insert_match_adapter_error_rate = options.insert_match_error_rate
                
        if options.merge_overlapping:
            if options.merged_output is None:
                parser.warn("--merge-output is not set; merged reads will be discarded")
            if options.merge_error_rate is None:
                options.merge_error_rate = options.error_rate or 0.2
        
        if options.mirna:
            if options.adapter is None and options.front is None and options.anywhere is None:
                options.adapter = ['TGGAATTCTCGG'] # illumina small RNA adapter
            if options.quality_cutoff is None:
                options.quality_cutoff = (20, 20)
            if options.minimum_length is None:
                options.minimum_length = 16
            if options.error_rate is None:
                options.error_rate = 0.12
        elif options.bisulfite:
            # TODO: set default adapter sequences
            # Jury is out on whether quality trimming helps. For aligners like
            # bwameth, it actually leads to worse results.
            #if options.quality_cutoff is None:
            #    options.quality_cutoff = "20,20"
            if options.bisulfite == "swift" and paired != "both":
                parser.error("Swift trimming is only compatible with paired-end reads")
            if options.bisulfite not in ("rrbs", "non-directional", "truseq", "epignome", "swift"):
                def parse_bisulfite_params(r):
                    try:
                        parts = [int(p) for p in r.split(",")]
                        assert len(parts) == 4
                        if parts[0] <= 0 and parts[1] <= 0:
                            return None
                        return dict(zip(
                            ("lengths", "count_trimmed", "only_trimmed"),
                            ((parts[0], -1 * parts[1]), (False,True)[parts[2]], (False,True)[parts[3]])
                        ))
                    except:
                        parser.error("Invalidate format for bisulfite parameters")
                
                temp = [parse_bisulfite_params(r) for r in options.bisulfite.split(";")]
                if paired == "both" and len(temp) == 1:
                    temp = [temp[0], temp[0]]
                elif paired != "both" and len(temp) > 1:
                    parser.error("Too many bisulfite parameters for single-end reads")
                options.bisulfite = temp
        
        if options.quality_cutoff:
            if all(c <= 0 for c in options.quality_cutoff):
                options.quality_cutoff = None
            elif len(options.quality_cutoff) == 1:
                options.quality_cutoff = [0] + options.quality_cutoff
        
        if options.pair_filter is None:
            options.pair_filter = 'any'

        if (options.discard_trimmed or options.discard_untrimmed) and (options.untrimmed_output is not None):
            parser.error("Only one of the --discard-trimmed, --discard-untrimmed "
                "and --untrimmed-output options can be used at the same time.")
        
        if options.output is not None and '{name}' in options.output:
            if options.discard_trimmed:
                parser.error("Do not use --discard-trimmed when demultiplexing.")
            if paired:
                parser.error("Demultiplexing not supported for paired-end files, yet.")
        
        if options.maq:
            options.colorspace = True
            options.double_encode = True
            options.trim_primer = True
            options.suffix = "/1"
        
        if options.strip_f3 or options.maq:
            options.strip_suffix.append('_F3')
        
        if options.zero_cap is None:
            options.zero_cap = options.colorspace
        
        if options.colorspace:
            if options.anywhere:
                parser.error("Using --anywhere with colorspace reads is currently not supported (if you "
                    "think this may be useful, contact the author).")
            if options.match_read_wildcards:
                parser.error('IUPAC wildcards not supported in colorspace')
            options.match_adapter_wildcards = False
        else:
            if options.trim_primer:
                parser.error("Trimming the primer makes only sense in colorspace.")
            if options.double_encode:
                parser.error("Double-encoding makes only sense in colorspace.")
                
        if options.error_rate is None:
            options.error_rate = 0.1
        
        if options.cut:
            if len(options.cut) > 2:
                parser.error("You cannot remove bases from more than two ends.")
            if len(options.cut) == 2 and options.cut[0] * options.cut[1] > 0:
                parser.error("You cannot remove bases from the same end twice.")
        
        if options.cut_min:
            if len(options.cut_min) > 2:
                parser.error("You cannot remove bases from more than two ends.")
            if len(options.cut_min) == 2 and options.cut_min[0] * options.cut_min[1] > 0:
                parser.error("You cannot remove bases from the same end twice.")
        
        if paired == 'both' and options.cut2:
            if len(options.cut2) > 2:
                parser.error("You cannot remove bases from more than two ends.")
            if len(options.cut2) == 2 and options.cut2[0] * options.cut2[1] > 0:
                parser.error("You cannot remove bases from the same end twice.")

        if paired == 'both' and options.cut_min2:
            if len(options.cut_min2) > 2:
                parser.error("You cannot remove bases from more than two ends.")
            if len(options.cut_min2) == 2 and options.cut_min2[0] * options.cut_min2[1] > 0:
                parser.error("You cannot remove bases from the same end twice.")
        
        if options.threads is not None:
            if options.debug:
                parser.error("Cannot use debug mode with multiple threads")
            
            threads = options.threads
            if threads <= 0:
                threads = cpu_count()
            elif threads == 1:
                parser.error("--threads must be >= 2")
            options.threads = threads
            
            if options.compression is None:
                # Our tests show that with 8 or more threads, worker compression is
                # more efficient.
                if options.writer_process and 2 < threads < 8:
                    from atropos.compression import can_use_system_compression
                    options.compression = "writer" if can_use_system_compression() else "worker"
                else:
                    options.compression = "worker"
            elif options.compression == "writer":
                if not options.writer_process:
                    parser.error("Writer compression and --no-writer-process are mutually exclusive")
                elif threads == 2:
                    logging.getLogger.warn("Writer compression requires > 2 threads; using worker compression instead")
                    options.compression = "worker"
            
            # Set queue sizes if necessary.
            # If we are using writer compression, the back-up will be in the result queue,
            # otherwise it will be in the read queue.
            if options.read_queue_size is None:
                options.read_queue_size = threads * (100 if options.compression == "writer" else 1000)
            elif options.read_queue_size > 0:
                assert options.read_queue_size >= threads
        
            if options.result_queue_size is None:
                options.result_queue_size = threads * (100 if options.compression == "worker" else 1000)
            elif options.result_queue_size > 0:
                assert options.result_queue_size >= threads

COMMANDS['trim'] = TrimCommand

class DetectCommand(Command):
    """detect adapter and other contaminant sequences."""
    name = "detect"
    usage = "atropos -se input.fastq detect\natropos -pe1 in1.fq -pe2 in2.fq detect"
    description = "Detect adapter sequences directly from read sequences."
    
    def add_command_options(self):
        parser = self.parser
        parser.set_defaults(max_reads=10000)
        parser.add_argument(
            "-d",
            "--detector",
            choices=('known', 'heuristic', 'khmer'), default=None,
            help="Which detector to use. (automatically choose based on other options)")
        parser.add_argument(
            "-o",
            "--output",
            type=writeable_file, default=STDOUT, metavar="FILE",
            help="File in which to write the summary of detected adapters. (stdout)")
        parser.add_argument(
            "-k",
            "--kmer-size",
            type=positive(), default=12,
            help="Size of k-mer used to scan reads for adapter sequences. (12)")
        parser.add_argument(
            "-m",
            "--max-adapters",
            type=positive(), default=None,
            help="The maximum number of candidate adapters to report. (report all)")
        
        group = parser.add_argument_group("Contaminants")
        group.add_argument(
            "-i",
            "--include-contaminants",
            choices=('all','known','unknown'), default='all',
            help="What conaminants to search for: all, only known adapters/contaminants ('known'), "
                 "or only unknown contaminants ('unknown'). (all)")
        group.add_argument(
            "-x",
            "--known-contaminant",
            action="append", dest='known_adapter', default=None,
            help="Pass known contaminants in on the commandline as 'name=sequence'. "
                 "Can be specified multiple times.")
        group.add_argument(
            "-F",
            "--known-contaminants-file",
            type=readable_url, action="append", dest='known_adapters_file', default=None,
            help="Points to FASTA File or URL with known contaminants.")
        group.add_argument(
            "--no-default-contaminants",
            action="store_false", dest="default_adapters", default=True,
            help="Don't fetch the default contaminant list (which is currently stored as a "
                "GitHub gist).")
        group.add_argument(
            "--contaminant-cache-file",
            type=readwriteable_file, dest='adapter_cache_file', default='.adapters',
            help="File where known contaminant sequences will be cached, unless "
                "--no-cache-contaminants is set.")
        group.add_argument(
            "--no-cache-contaminants",
            action="store_false", dest="cache_adapters", default=True,
            help="Don't cache contaminant list as '.contaminants' in the working directory.")

COMMANDS['detect'] = DetectCommand

class ErrorCommand(Command):
    """estimate the sequencing error rate."""
    name = "error"
    usage = "atropos -se input.fastq error\natropos -pe in1.fq -pe2 in2.fq error"
    description = """
Estimate the error rate from base qualities. This can help to
determine the quality of your data, and to decide the value for
the max error rate (-e) parameter. Normal error for an Illumina
experiment is around 1% or less. We recommend setting -e to
10 x the empirical error rate."""
    
    def add_command_options(self):
        parser = self.parser
        parser.set_defaults(max_reads=10000)
        parser.add_argument(
            "-o",
            "--output",
            type=writeable_file, default=STDOUT,
            help="File in which to write the summary of the estimated error rates. (stdout)")

COMMANDS['error'] = ErrorCommand

# Main

def main(cmdlineargs=None):
    """
    Main function that evaluates command-line parameters.
    
    :param cmdlineargs: iterable of command line arguments; `None` is equivalent to
    `sys.argv[1:]`
    """
    args = cmdlineargs or sys.argv[1:]
    if len(args) == 0 or any(h in args for h in ('-h', '--help')):
        print_subcommands()
        return
    
    if args[0][0] == '-':
        command_name = "trim"
    else:
        command_name = args[0]
        del args[0]
    
    command_class = COMMANDS[command_name]
    command = command_class(args)
    command.execute()

def print_subcommands():
    print("Atropos version {}\n".format(__version__))
    print("usage: atropos <command> [options]\n")
    print("commands:")
    for command_class in COMMANDS.values():
        print("  {}: {}".format(command_class.name, command_class.__doc__))
    print("\noptional arguments:\n  -h, --help  show this help message and exit")
    print("""
Use "atropos <command> --help" to see all options for a specific command.
See http://atropos.readthedocs.org/ for full documentation.

Atropos is a fork of Cutadapt 1.10 (
https://github.com/marcelm/cutadapt/tree/2f3cc0717aa9ff1e0326ea6bcb36b712950d4999)
by John Didion, "Atropos: sensitive, specific, and speedy trimming of NGS reads,"
in prep.

Cutadapt (https://github.com/marcelm/cutadapt) was developed by Marcel Martin,
"Cutadapt Removes Adapter Sequences From High-Throughput Sequencing Reads,"
EMBnet Journal, 2011, 17(1):10-12.
""")

if __name__ == '__main__':
    main()



# coding: utf-8
from __future__ import print_function, division, absolute_import

from atropos.align import locate, compare_prefixes, compare_suffixes, Aligner, InsertAligner
from atropos.util import RandomMatchProbability
from atropos.adapters import BACK


class TestAligner():
    def test(self):
        reference = 'CTCCAGCTTAGACATATC'
        aligner = Aligner(reference, 0.1, flags=BACK)
        aligner.locate('CC')

    def test_100_percent_error_rate(self):
        reference = 'GCTTAGACATATC'
        aligner = Aligner(reference, 1.0, flags=BACK)
        aligner.locate('CAA')


def test_polya():
    s = 'AAAAAAAAAAAAAAAAA'
    t = 'ACAGAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA'
    result = locate(s, t, 0.0, BACK)
    #start_s, stop_s, start_t, stop_t, matches, cost = result
    assert result == (0, len(s), 4, 4 + len(s), len(s), 0)


# Sequences with IUPAC wildcards
# R=A|G, Y=C|T, S=G|C, W=A|T, K=G|T, M=A|C, B=C|G|T, D=A|G|T, H=A|C|T, V=A|C|G,
# N=A|C|G|T, X={}
WILDCARD_SEQUENCES = [
    'CCCATTGATC',  # original sequence without wildcards
    'CCCRTTRATC',  # R=A|G
    'YCCATYGATC',  # Y=C|T
    'CSSATTSATC',  # S=G|C
    'CCCWWWGATC',  # W=A|T
    'CCCATKKATC',  # K=G|T
    'CCMATTGMTC',  # M=A|C
    'BCCATTBABC',  # B=C|G|T
    'BCCATTBABC',  # B
    'CCCDTTDADC',  # D=A|G|T
    'CHCATHGATC',  # H=A|C|T
    'CVCVTTVATC',  # V=A|C|G
    'CCNATNGATC',  # N=A|C|G|T
    'CCCNTTNATC',  # N
#   'CCCXTTXATC',  # X
]


def test_compare_prefixes():
    assert compare_prefixes('AAXAA', 'AAAAATTTTTTTTT') == (0, 5, 0, 5, 4, 1)
    assert compare_prefixes('AANAA', 'AACAATTTTTTTTT', wildcard_ref=True) == (0, 5, 0, 5, 5, 0)
    assert compare_prefixes('AANAA', 'AACAATTTTTTTTT', wildcard_ref=True) == (0, 5, 0, 5, 5, 0)
    assert compare_prefixes('XAAAAA', 'AAAAATTTTTTTTT') == (0, 6, 0, 6, 4, 2)

    a = WILDCARD_SEQUENCES[0]
    for s in WILDCARD_SEQUENCES:
        r = s + 'GCCAGGGTTGATTCGGCTGATCTGGCCG'
        result = compare_prefixes(a, r, wildcard_query=True)
        assert result == (0, 10, 0, 10, 10, 0), result

        result = compare_prefixes(r, a, wildcard_ref=True)
        assert result == (0, 10, 0, 10, 10, 0)

    for s in WILDCARD_SEQUENCES:
        for t in WILDCARD_SEQUENCES:
            r = s + 'GCCAGGG'
            result = compare_prefixes(s, r, )
            assert result == (0, 10, 0, 10, 10, 0)

            result = compare_prefixes(r, s, wildcard_ref=True, wildcard_query=True)
            assert result == (0, 10, 0, 10, 10, 0)

    r = WILDCARD_SEQUENCES[0] + 'GCCAGG'
    for wildc_ref in (False, True):
        for wildc_query in (False, True):
            result = compare_prefixes('CCCXTTXATC', r, wildcard_ref=wildc_ref, wildcard_query=wildc_query)
            assert result == (0, 10, 0, 10, 8, 2)


def test_compare_suffixes():
    assert compare_suffixes('AAXAA', 'TTTTTTTAAAAA') == (0, 5, 7, 12, 4, 1)
    assert compare_suffixes('AANAA', 'TTTTTTTAACAA', wildcard_ref=True) == (0, 5, 7, 12, 5, 0)
    assert compare_suffixes('AANAA', 'TTTTTTTAACAA', wildcard_ref=True) == (0, 5, 7, 12, 5, 0)
    assert compare_suffixes('AAAAAX', 'TTTTTTTAAAAA') == (0, 6, 6, 12, 4, 2)


def test_wildcards_in_adapter():
    r = 'CATCTGTCC' + WILDCARD_SEQUENCES[0] + 'GCCAGGGTTGATTCGGCTGATCTGGCCG'
    for a in WILDCARD_SEQUENCES:
        result = locate(a, r, 0.0, BACK, wildcard_ref=True)
        assert result == (0, 10, 9, 19, 10, 0), result

    a = 'CCCXTTXATC'
    result = locate(a, r, 0.0, BACK, wildcard_ref=True)
    assert result is None


def test_wildcards_in_read():
    a = WILDCARD_SEQUENCES[0]
    for s in WILDCARD_SEQUENCES:
        r = 'CATCTGTCC' + s + 'GCCAGGGTTGATTCGGCTGATCTGGCCG'
        result = locate(a, r, 0.0, BACK, wildcard_query=True)
        if 'X' in s:
            assert result is None
        else:
            assert result == (0, 10, 9, 19, 10, 0), result


def test_wildcards_in_both():
    for a in WILDCARD_SEQUENCES:
        for s in WILDCARD_SEQUENCES:
            if 'X' in s or 'X' in a:
                continue
            r = 'CATCTGTCC' + s + 'GCCAGGGTTGATTCGGCTGATCTGGCCG'
            result = locate(a, r, 0.0, BACK, wildcard_ref=True, wildcard_query=True)
            assert result == (0, 10, 9, 19, 10, 0), result


def test_no_match():
    a = locate('CTGATCTGGCCG', 'AAAAGGG', 0.1, BACK)
    assert a is None, a

import math
from .utils import approx_equal

def test_factorial():
    f = RandomMatchProbability()
    # simple test
    assert f.factorial(0) == 1
    assert f.factorial(1) == 1
    assert f.factorial(3) == 6
    assert int(f.factorial(27)) == int(math.factorial(27))
    # test big number
    assert int(f.factorial(150)) == int(math.factorial(150))

def test_match_probability():
    a = InsertAligner('TTAGACATAT', 'CAGTGGAGTA')
    k = 3
    n = 5
    i3 = (120 / (6 * 2)) * (0.25 ** 3) * (0.75 ** 2)
    i4 = (120 / 24) * (0.25 ** 4) * 0.75
    i5 = 0.25 ** 5
    assert approx_equal(a.match_probability(k, n), i3 + i4 + i5, 0.0001)

def test_insert_align():
    a1_seq = 'TTAGACATATGG'
    a2_seq = 'CAGTGGAGTATA'
    aligner = InsertAligner(a1_seq, a2_seq)
    r1 = 'AGTCGAGCCCATTGCAGACT' + a1_seq[0:10]
    r2 = 'AGTCTGCAATGGGCTCGACT' + a2_seq[0:10]
    insert_match, match1, match2 = aligner.match_insert(r1, r2)
    assert match1.rstart == 20
    assert match1.length == 10
    assert match2.rstart == 20
    assert match2.length == 10
    
def test_short_adapter_overlap():
    a1_seq = 'TTAGACATAT'
    a2_seq = 'CAGTGGAGTA'
    seq1 = 'GACAGGCCGTTTGAATGTTGACGGGATGTT'
    seq2 = 'CATCCCGTCAACATTCAAACGGCCTGTCCA'
    aligner = InsertAligner(a1_seq, a2_seq)
    insert_match, match1, match2 = aligner.match_insert(seq1, seq2)
    assert match1.rstart == 28
    assert match1.length == 2
    assert match2.rstart == 28
    assert match2.length == 2

def test_multi_aligner_no_mismatches():
    from atropos._align import MultiAligner
    a = MultiAligner(max_error_rate=0, min_overlap=3)
    matches = a.locate('AGAGATCAGATGACAGATC', 'GATCA')
    assert len(matches) == 2
    
    matches.sort(key=lambda x: x[4], reverse=True)
    
    assert matches[0][0] == 3
    assert matches[0][1] == 8
    assert matches[0][2] == 0
    assert matches[0][3] == 5
    assert matches[0][4] == 5
    assert matches[0][5] == 0
    
    assert matches[1][0] == 15
    assert matches[1][1] == 19
    assert matches[1][2] == 0
    assert matches[1][3] == 4
    assert matches[1][4] == 4
    assert matches[1][5] == 0

def test_multi_aligner_with_mismatches():
    from atropos._align import MultiAligner
    a = MultiAligner(max_error_rate=0.1, min_overlap=10)
    matches = a.locate('GATATCAGATGACAGATCAGAGATCAGAT', 'GAGATCAGATGA')
    
    assert len(matches) == 2
    
    matches.sort(key=lambda x: x[5])
    
    assert matches[0][0] == 19
    assert matches[0][1] == 29
    assert matches[0][2] == 0
    assert matches[0][3] == 10
    assert matches[0][4] == 10
    assert matches[0][5] == 0
    
    assert matches[1][0] == 0
    assert matches[1][1] == 12
    assert matches[1][2] == 0
    assert matches[1][3] == 12
    assert matches[1][4] == 11
    assert matches[1][5] == 1



import logging
import time
from .util import MAGNITUDE

def magnitude_formatter(magnitude):
    suffix = ""
    if magnitude is None:
        div = 1.0
    else:
        div = float(MAGNITUDE[magnitude.upper()])
        suffix = magnitude
    return lambda val: "{:.1f} {}".format(val / div, suffix)

def create_progress_reader(reader, progress_type="bar", batch_size=1, max_items=None,
                           counter_magnitude="M", **kwargs):
    mag_format = magnitude_formatter(counter_magnitude)
    
    if progress_type == "msg":
        return ProgressMessageReader(reader, batch_size, max_items=max_items, mag_format=mag_format, **kwargs)
            
    try:
        return create_progressbar_reader(reader, max_items, mag_format, **kwargs)
    except:
        pass
    
    try:
        return create_tqdm_reader(reader, max_items, **kwargs)
    except:
        pass
    
    logging.getLogger().warn("No progress bar library available")
    return reader

class ProgressMessageReader(object):
    def __init__(self, iterable, batch_size, interval=1000000, max_items=None, mag_format=None):
        self.iterable = iterable
        self.batch_size = batch_size
        self.interval = interval
        self.ctr = 0
        self.mag_format = mag_format
        if max_items:
            if mag_format:
                max_items = mag_format(max_items)
            else:
                max_items = str(max_items)
            self.msg = "Read {0}/" + max_items + " records in {1:.1f} seconds"
        else:
            self.msg = "Read {0} records in {1:.1f} seconds"
        
    def __next__(self):
        value = next(self.iterable)
        if value:
            self.ctr += value[0]
            if self.ctr % self.interval < self.batch_size:
                now = time.time()
                ctr = self.mag_format(self.ctr) if self.mag_format else self.ctr
                logging.getLogger().info(self.msg.format(ctr, now - self.start))
        return value
    
    next = __next__
    def __iter__(self):
        self.start = time.time()
        return self
    
    def close(self):
        logging.getLogger().info("Read a total of {} records".format(self.ctr))
        self.iterable.close()

def create_progressbar_reader(reader, max_reads=None, mag_format=None):
    import progressbar
    import progressbar.widgets
    import math

    class ProgressBarReader(progressbar.ProgressBar):
        def __init__(self, iterable, widgets, max_value=None):
            super(ProgressBarReader, self).__init__(
                widgets=widgets, max_value=max_value or progressbar.UnknownLength)
            self._iterable = iterable
            self.done = False
        
        def __next__(self):
            try:
                value = next(self._iterable)
                if self.start_time is None:
                    self.start()
                self.update(self.value + value[0])
                return value
            except StopIteration:
                self.close()
                raise
        
        def close(self):
            if not self.done:
                self.finish()
                self.done = True
            try:
                self._iterable.close()
            except:
                pass
    
    class MagCounter(progressbar.widgets.WidgetBase):
        def __init__(self, mag_format):
            self._format = mag_format
    
        def __call__(self, progress, data):
            return self._format(data["value"])
        
    if max_reads:
        reader = ProgressBarReader(reader, [
            MagCounter(mag_format), " Reads (", progressbar.Percentage(), ") ",
            progressbar.Timer(), " ", progressbar.Bar(), progressbar.AdaptiveETA()
        ], max_reads)
    else:
        reader = ProgressBarReader(reader, [
            MagCounter(mag_format), " Reads", progressbar.Timer(),
            progressbar.AnimatedMarker()
        ])
    
    return reader

def create_tqdm_reader(reader, max_reads=None):
    import tqdm
    return tqdm.tqdm(reader)



# Estimate the empircal error rate

from .seqio import FastqReader
from .util import enumerate_range

def qual2prob(qchar):
    q = ord(qchar) - 33
    return 10 ** (-q / 10)

class ErrorEstimator(object):
    def __init__(self):
        self.total_qual = 0.0
        self.total_len = 0
    
    def consume_all_batches(self, batch_iterator):
        for batch_num, batch in batch_iterator:
            for read in batch:
                self.consume(read)
    
    def consume(self, read):
        self.total_qual += sum(qual2prob(qchar) for qchar in read.qualities)
        self.total_len += len(read.qualities)
    
    def estimate(self):
        return self.total_qual / self.total_len
    
    def summarize(self, outstream, name=None):
        print("", file=outstream)
        if name is not None:
            header = "File: {}".format(name)
            print(header)
            print('-' * len(header), file=outstream)
        print("Error rate: {:.2%}".format(self.estimate()))

class PairedErrorEstimator(object):
    def __init__(self):
        self.e1 = ErrorEstimator()
        self.e2 = ErrorEstimator()
    
    def consume_all_batches(self, batch_iterator):
        for batch_num, batch in batch_iterator:
            for read1, read2 in batch:
                self.e1.consume(read1)
                self.e2.consume(read2)
    
    def estimate(self):
        return (self.e1.estimate(), self.e2.estimate())
    
    def summarize(self, outstream, names=None):
        print("", file=outstream)
        if names:
            header = "File 1: {}".format(names[0])
            print(header)
            print('-' * len(header), file=outstream)
        print("Error rate: {:.2%}\n".format(self.e1.estimate()))
        
        if names:
            header = "File 2: {}".format(names[1])
            print(header)
            print('-' * len(header), file=outstream)
        print("Error rate: {:.2%}\n".format(self.e2.estimate()))
        
        print("Overall error rate: {:.2%}".format(
            (self.e1.total_qual + self.e2.total_qual) / (self.e1.total_len + self.e2.total_len)))



import math

base_complements = {
    'A' : 'T',
    'C' : 'G',
    'R' : 'Y',
    'S' : 'S',
    'W' : 'W',
    'K' : 'M',
    'B' : 'V',
    'D' : 'H',
    'N' : 'N'
}
for k,v in list(base_complements.items()):
    base_complements[v] = k
    base_complements[k.lower()] = v.lower()
    base_complements[v.lower()] = k.lower()

MAGNITUDE = dict(
    G=1E9,
    M=1E6,
    K=1E3
)

log2 = math.log(2)

def complement(seq):
    return "".join(base_complements[base] for base in seq)

def reverse_complement(seq):
    return "".join(base_complements[base] for base in reversed(seq))

def sequence_complexity(seq):
    """A simple measure of sequence complexity"""
    seq = seq.upper()
    seqlen = float(len(seq))
    term = 0
    for base in ('A','C','G','T'):
        c = seq.count(base)
        if c > 0:
            d = c / seqlen
            term += d * math.log(d) / log2
    return -term

def enumerate_range(collection, start, end):
    'Generates an indexed series:  (0,coll[0]), (1,coll[1]) ...'
    i = start
    it = iter(collection)
    while i < end:
        yield (i, next(it))
        i += 1

def mean(data):
    return sum(data) / len(data)

def median(data):
    """
    Median function borrowed from python statistics module, and sped up by
    in-place sorting of the array.
    """
    n = len(data)
    if n == 0:
        raise Exception("no median for empty data")
    
    data.sort()
    
    i = n // 2
    if n % 2 == 1:
        return data[i]
    else:
        return (data[i - 1] + data[i]) / 2

class RandomMatchProbability(object):
    """
    Class for computing random match probability for DNA sequences
    based on binomial expectation. Maintains a cache of factorials
    to speed computation.
    """
    def __init__(self, init_size=150):
        self.cache = {}
        self.factorials = [1] * init_size
        self.max_n = 1
        self.cur_array_size = init_size
    
    def __call__(self, matches, size, p1=0.25, p2=0.75):
        # First see if we have the result in the cache
        key = (matches, size)
        p = self.cache.get(key, None)
        if p:
            return p
        
        # When there are no mismatches, the probability is
        # just that of observing a specific sequence of the
        # given length by chance.
        if matches == size:
            p = p1 ** matches
        
        else:
            nfac = self.factorial(size)
            p = 0.0
            for i in range(matches, size+1):
                j = size - i
                p += (
                    (p2 ** j) *
                    (p1 ** i) *
                    nfac /
                    self.factorial(i) /
                    self.factorial(j)
                )

        self.cache[key] = p
        return p
    
    def factorial(self, n):
        if n > self.max_n:
            self._fill_upto(n)
        return self.factorials[n]

    def _fill_upto(self, n):
        if n >= self.cur_array_size:
            extension_size = n - self.cur_array_size + 1
            self.factorials += [1] * extension_size
        i = self.max_n
        next_i = i + 1
        while i < n:
            self.factorials[next_i] = next_i * self.factorials[i]
            i = next_i
            next_i += 1
        self.max_n = i
